'''
Created on 2016/4/11

:author: hubo
'''
from vlcp.event.runnable import RoutineContainer, GeneratorExit_
from vlcp.server.module import call_api, send_api
from uuid import uuid1
from vlcp.event.event import Event, withIndices, M_
from vlcp.utils.dataobject import multiwaitif
from vlcp.protocol.openflow.openflow import OpenflowErrorResultException
from namedstruct.namedstruct import dump
import json
import logging
from contextlib import closing

@withIndices('updater', 'type')
class FlowUpdaterNotification(Event):
    """
    These notifications are edge triggered - they work with identifiers
    and set
    """
    STARTWALK = 'startwalk'
    DATAUPDATED = 'dataupdated'
    FLOWUPDATE = 'flowupdate'

class FlowUpdater(RoutineContainer):
    def __init__(self, connection, initialkeys = (), requestid = None, logger = None):
        """
        Retrieve data objects from ObjectDB and use them to generate flows
        
        The module starts ObjectDB.walk from initial keys and walkers. After the
        walk completes, the retrieved data objects are used by `updateflow()` to
        generate flows and send them to the OpenFlow connection. When the retrieved
        objects are updated, FlowUpdater either restart the walk process (re-walk)
        or directly call another `updateflow()`, according to the objects that are updated.
        
        A subclass should re-initialize `self._initialkeys` and `self._walkerdict`
        before `main()` coroutine starts to customize the process.
        
        `updateflow()` is guaranteed for no re-entrance i.e. will not be called until
        the last call returns. Multiple changes may be merged into the same call.
        
        :param connection: OpenFlow connection
        
        :param initialkeys: DEPRECATED The key list that triggers a re-walk
        
        :param requestid: request id to retrieve data objects from ObjectDB
        
        :param logger: inherit a logger from a module
        """
        RoutineContainer.__init__(self, connection.scheduler)
        # When keys in this sequence are updated, start re-walk
        self._initialkeys = initialkeys
        self._connection = connection
        # Walker dict, will be re-initialized by a subclass
        self._walkerdict = {}
        # Walk result
        self._savedkeys = ()
        self._savedresult = []
        # Update notification (with a cache for synchronize)
        self._updatedset = set()
        self._updatedset2 = set()
        if not logger:
            self._logger = logging.getLogger(__name__ + '.FlowUpdater')
        else:
            self._logger = logger
        if requestid is None:
            self._requstid = str(uuid1())
        else:
            self._requstid = requestid
        self._requestindex = 0
        # Detect data object updates
        self._dataupdateroutine = None
        # Calling updateflow() in the same routine to prevent a re-entrance
        self._flowupdateroutine = None

    def reset_initialkeys(self, keys, values):
        """
        Callback after walk complete, can be used to update `self._initialkeys`.
        """
        pass

    async def walkcomplete(self, keys, values):
        """
        Async callback after walk complete, before flow update
        """
        pass

    async def updateflow(self, connection, addvalues, removevalues, updatedvalues):
        """
        Update flow callback. When data objects are updated (either by a re-walk
        or by a direct update), this method is called with the modification, after
        the last `updateflow()` call ends.
        """
        pass

    def shouldupdate(self, newvalues, updatedvalues):
        """
        Callback when upate. Rewrite this method to ignore some updates.
        
        If this callback returns False, the update is ignored.
        """
        return True

    async def restart_walk(self):
        """
        Force a re-walk
        """
        if not self._restartwalk:
            self._restartwalk = True
            await self.wait_for_send(FlowUpdaterNotification(self, FlowUpdaterNotification.STARTWALK))

    async def _dataobject_update_detect(self, _initialkeys, _savedresult):
        """
        Coroutine that wait for retrieved value update notification
        """
        def expr(newvalues, updatedvalues):
            if any(v.getkey() in _initialkeys for v in updatedvalues if v is not None):
                return True
            else:
                return self.shouldupdate(newvalues, updatedvalues)
        while True:
            updatedvalues, _ = await multiwaitif(_savedresult, self, expr, True)
            if not self._updatedset:
                self.scheduler.emergesend(FlowUpdaterNotification(self, FlowUpdaterNotification.DATAUPDATED))
            self._updatedset.update(updatedvalues)

    def updateobjects(self, updatedvalues):
        """
        Force a update notification on specified objects, even if they are not actually updated
        in ObjectDB
        """
        if not self._updatedset:
            self.scheduler.emergesend(FlowUpdaterNotification(self, FlowUpdaterNotification.DATAUPDATED))
        self._updatedset.update(set(updatedvalues).intersection(self._savedresult))

    async def _flowupdater(self):
        """
        Coroutine calling `updateflow()`
        """
        lastresult = set(v for v in self._savedresult if v is not None and not v.isdeleted())
        flowupdate = FlowUpdaterNotification.createMatcher(self, FlowUpdaterNotification.FLOWUPDATE)
        while True:
            currentresult = [v for v in self._savedresult if v is not None and not v.isdeleted()]
            # Calculating differences
            additems = []
            updateditems = []
            updatedset2 = self._updatedset2
            for v in currentresult:
                if v not in lastresult:
                    additems.append(v)
                else:
                    lastresult.remove(v)
                    if v in updatedset2:
                        # Updated
                        updateditems.append(v)
            removeitems = lastresult
            self._updatedset2.clear()
            # Save current result for next difference
            lastresult = set(currentresult)
            if not additems and not removeitems and not updateditems:
                await flowupdate
                continue
            await self.updateflow(self._connection, set(additems), removeitems, set(updateditems))

    async def main(self):
        """
        Main coroutine
        """
        try:
            lastkeys = set()
            dataupdate = FlowUpdaterNotification.createMatcher(self, FlowUpdaterNotification.DATAUPDATED)
            startwalk = FlowUpdaterNotification.createMatcher(self, FlowUpdaterNotification.STARTWALK)
            self.subroutine(self._flowupdater(), False, '_flowupdateroutine')
            # Cache updated objects
            presave_update = set()
            while True:
                self._restartwalk = False
                presave_update.update(self._updatedset)
                self._updatedset.clear()
                _initialkeys = set(self._initialkeys)
                try:
                    walk_result = await call_api(self, 'objectdb', 'walk',
                                                        {'keys': self._initialkeys, 'walkerdict': self._walkerdict,
                                                         'requestid': (self._requstid, self._requestindex)})
                except Exception:
                    self._logger.warning("Flow updater %r walk step failed, conn = %r", self, self._connection,
                                         exc_info=True)
                    # Cleanup
                    await call_api(self, 'objectdb', 'unwatchall',
                                         {'requestid': (self._requstid, self._requestindex)})
                    await self.wait_with_timeout(2)
                    self._requestindex += 1
                if self._restartwalk:
                    continue
                if self._updatedset:
                    if any(v.getkey() in _initialkeys for v in self._updatedset):
                        # During walk, there are other initial keys that are updated
                        # To make sure we get the latest result, restart the walk
                        continue
                lastkeys = set(self._savedkeys)
                _savedkeys, _savedresult = walk_result
                removekeys = tuple(lastkeys.difference(_savedkeys))
                self.reset_initialkeys(_savedkeys, _savedresult)
                _initialkeys = set(self._initialkeys)
                if self._dataupdateroutine:
                    self.terminate(self._dataupdateroutine)
                # Start detecting updates
                self.subroutine(self._dataobject_update_detect(_initialkeys, _savedresult), False, "_dataupdateroutine")
                # Set the updates back (potentially merged with newly updated objects)
                self._updatedset.update(v for v in presave_update)
                presave_update.clear()
                await self.walkcomplete(_savedkeys, _savedresult)
                if removekeys:
                    await call_api(self, 'objectdb', 'munwatch', {'keys': removekeys,
                                                                  'requestid': (self._requstid, self._requestindex)})
                # Transfer updated objects to updatedset2 before a flow update notification
                # This helps to make `walkcomplete` executes before `updateflow`
                #
                # But notice that since there is only a single data object copy in all the program,
                # it is impossible to hide the change completely during `updateflow`
                self._updatedset2.update(self._updatedset)
                self._updatedset.clear()
                self._savedkeys = _savedkeys
                self._savedresult = _savedresult
                await self.wait_for_send(FlowUpdaterNotification(self, FlowUpdaterNotification.FLOWUPDATE))
                while not self._restartwalk:
                    if self._updatedset:
                        if any(v.getkey() in _initialkeys for v in self._updatedset):
                            break
                        else:
                            self._updatedset2.update(self._updatedset)
                            self._updatedset.clear()
                            self.scheduler.emergesend(FlowUpdaterNotification(self, FlowUpdaterNotification.FLOWUPDATE))
                    await M_(dataupdate, startwalk)
        except Exception:
            self._logger.exception("Flow updater %r stops update by an exception, conn = %r", self, self._connection)
            raise
        finally:
            self.subroutine(send_api(self, 'objectdb', 'unwatchall', {'requestid': (self._requstid, self._requestindex)}),
                            False)
            if self._flowupdateroutine:
                self.terminate(self._flowupdateroutine)
                self._flowupdateroutine = None
            if self._dataupdateroutine:
                self.terminate(self._dataupdateroutine)
                self._dataupdateroutine = None

    async def execute_commands(self, conn, cmds):
        if cmds:
            try:
                _, openflow_replydict = await conn.protocol.batch(cmds, conn, self)
            except OpenflowErrorResultException as exc:
                self._logger.warning("Some Openflow commands return error result on connection %r, will ignore and continue.\n"
                                             "Details:\n%s", conn,
                                             "\n".join("REQUEST = \n%s\nERRORS = \n%s\n" % (json.dumps(dump(k, tostr=True), indent=2),
                                                                                            json.dumps(dump(v, tostr=True), indent=2))
                                                       for k,v in exc.result[1].items()))



'''
Created on 2015/06/02

:author: hubo
'''
from __future__ import print_function, absolute_import, division 

from collections import deque
from .matchtree import MatchTree
from .event import Event, withIndices
from bisect import bisect_left
from heapq import heappush, heappop


@withIndices('queue')
class QueueCanWriteEvent(Event):
    def __init__(self, queue, **kwargs):
        Event.__init__(self, queue, **kwargs)
        self.queue = queue

    def is_valid(self):
        if self.queue.canAppend():
            return True
        else:
            self.queue.isWaited = True
            return False


@withIndices('queue')
class QueueIsEmptyEvent(Event):
    pass


@withIndices('key', 'newonly', 'firstonly')
class AutoClassQueueCanWriteEvent(QueueCanWriteEvent):
    def __init__(self, queue, key, newonly, firstonly):
        Event.__init__(self, queue, key, newonly, firstonly)
        self.queue = queue
        self.key = key

    def is_valid(self):
        if self.queue.maxlength is not None:
            return None
        else:
            if self.queue.canAppendSubqueue(self.key):
                return True
            else:
                self.queue.waited.add((False, False, self.key))
                return False


class CBQueue(object):
    '''
    A multi-queue model with priority and balance.
    When first created, there is a default queue with priority 0. More sub-queues maybe created with addSubQueue.
    Each sub-queue is a CBQueue which accepts more sub-queues. Sub-queues are considered as black-box to the outer parent.
    '''
    class FifoQueue(object):
        '''
        A wrapper for a FIFO queue
        '''
        def __init__(self, parent = None, maxlength = None):
            self.queue = deque()
            self.parent = parent
            self.maxlength = maxlength
            self.blocked = False
            if self.maxlength is not None and self.maxlength <= 0:
                self.maxlength = 1
            self.isWaited = False
            self._matcher = QueueCanWriteEvent.createMatcher(self)
        def append(self, value, force = False):
            if not force and not self.canAppend():
                self.isWaited = True
                return self._matcher
            if self.parent is not None:
                m = self.parent.notifyAppend(self, force)
                if m is not None:
                    return m
            self.queue.append(value)
            return None
        def canAppend(self):
            return self.maxlength is None or len(self.queue) < self.maxlength
        def canPop(self):
            return self.queue and not self.blocked
        def pop(self):
            ret = self._pop()
            if self.parent is not None:
                pr = self.parent.notifyPop(self)
                ret[1].extend(pr[0])
                ret[2].extend(pr[1])
            return ret
        def _pop(self):
            if self.blocked:
                raise IndexError('pop from a blocked queue')
            ret = self.queue.popleft()
            if self.isWaited and self.canAppend():
                self.isWaited = False
                return (ret, [QueueCanWriteEvent(self)], [])
            else:
                return (ret, [], [])
        def clear(self):
            l = len(self)
            ret = self._clear()
            if self.parent is not None:
                pr = self.parent.notifyPop(self, l)
                ret[0].extend(pr[0])
                ret[1].extend(pr[1])
            return ret
        def _clear(self):
            if self.blocked:
                self.unblockall()
            self.queue.clear()
            if self.isWaited and self.canAppend():
                self.isWaited = False
                return ([QueueCanWriteEvent(self)], [])
            else:
                return ([], [])
        def __len__(self):
            return len(self.queue)
        def block(self, value):
            if self.parent is not None:
                self.parent.notifyAppend(self, True)
            self.queue.appendleft(value)
            if not self.blocked:
                self.blocked = True
                if self.parent is not None:
                    self.parent.notifyBlock(self, True)
        def unblock(self, value):
            if self.blocked:
                self.blocked = False
                if self.parent is not None:
                    self.parent.notifyBlock(self, False)
        def unblockall(self):
            if self.blocked:
                self.blocked = False
                if self.parent is not None:
                    self.parent.notifyBlock(self, False)
    class PriorityQueue(object):
        '''
        A queue with inner built priority. Event must have a "priority" property to use with this type of queue.
        For fail-safe, events without "priority" property have the lowest priority.
        
        NOTICE: different from the queue priority, the priority property is smaller-higher, and is not limited to integers.
        This allows datetime to be used as an increasing priority
        '''
        def __init__(self, parent = None, maxlength = None, key = 'priority'):
            # a heap
            self.queue = []
            self.deque = deque()
            self.parent = parent
            self.maxlength = maxlength
            self.blocks = set()
            if self.maxlength is not None and self.maxlength <= 0:
                self.maxlength = 1
            self.blocked = False
            self.isWaited = False
            self.key = key
            self._matcher = QueueCanWriteEvent.createMatcher(self)
        @classmethod
        def initHelper(cls, key = 'priority'):
            def initer(parent = None, maxlength = None):
                return cls(parent, maxlength, key)
            return initer
        def append(self, value, force = False):
            if not force and not self.canAppend():
                self.isWaited = True
                return self._matcher
            if self.parent is not None:
                m = self.parent.notifyAppend(self, force)
                if m is not None:
                    return m
            if hasattr(value, self.key):
                heappush(self.queue, (getattr(value, self.key), value))
                # a priority push may change the block status
                if self.blocked and not self.queue[0][1] in self.blocks:
                    self.blocked = False
                    if self.parent is not None:
                        self.parent.notifyBlock(self, False)
            else:
                self.deque.append(value)
            return None
        def canAppend(self):
            return self.maxlength is None or len(self.queue) + len(self.deque) < self.maxlength
        def canPop(self):
            return len(self.queue) + len(self.deque) > 0 and not self.blocked
        def pop(self):
            ret = self._pop()
            if self.parent is not None:
                pr = self.parent.notifyPop(self)
                ret[1].extend(pr[0])
                ret[2].extend(pr[1])
            return ret
        def _pop(self):
            if self.blocked:
                raise IndexError('pop from a blocked queue')
            if self.queue:
                ret = heappop(self.queue)[1]
            else:
                ret = self.deque.popleft()
            if self.queue:
                blocked = self.queue[0][1] in self.blocks
            elif self.deque:
                blocked = self.deque[0] in self.blocks
            else:
                blocked = False
            if self.blocked != blocked:
                self.blocked = blocked
                if self.parent is not None:
                    self.parent.notifyBlock(self, blocked)
            if self.isWaited and self.canAppend():
                self.isWaited = False
                return (ret, [QueueCanWriteEvent(self)], [])
            else:
                return (ret, [], [])
        def _clear(self):
            if self.blocks:
                self.unblockall()
            del self.queue[:]
            self.deque.clear()
            if self.isWaited and self.canAppend():
                self.isWaited = False
                return ([QueueCanWriteEvent(self)], [])
            else:
                return ([], [])
        def clear(self):
            l = len(self)
            ret = self._clear()
            if self.parent is not None:
                pr = self.parent.notifyPop(self, l)
                ret[0].extend(pr[0])
                ret[1].extend(pr[1])
            return ret
        def __len__(self):
            return len(self.queue) + len(self.deque)
        def block(self, value):
            self.blocks.add(value)
            if self.parent is not None:
                self.parent.notifyAppend(self, True)
            if hasattr(value, self.key):
                heappush(self.queue, (getattr(value, self.key), value))
            else:
                self.deque.appendleft(value)
            if self.queue:
                blocked = self.queue[0][1] in self.blocks
            elif self.deque:
                blocked = self.deque[0] in self.blocks
            else:
                blocked = False
            if self.blocked != blocked:
                self.blocked = blocked
                if self.parent is not None:
                    self.parent.notifyBlock(self, blocked)
        def unblock(self, value):
            self.blocks.remove(value)
            if self.queue:
                blocked = self.queue[0][1] in self.blocks
            elif self.deque:
                blocked = self.deque[0] in self.blocks
            else:
                blocked = False
            if self.blocked != blocked:
                self.blocked = blocked
                if self.parent is not None:
                    self.parent.notifyBlock(self, blocked)
        def unblockall(self):
            self.blocks.clear()
            blocked = False
            if self.blocked != blocked:
                self.blocked = blocked
                if self.parent is not None:
                    self.parent.notifyBlock(self, blocked)
    class MultiQueue(object):
        '''
        A multi-queue container, every queue in a multi-queue has the same priority, and is popped in turn.
        '''
        class CircleListNode(object):
            '''
            Circle link list
            '''
            def __init__(self, value):
                self.prev = self
                self.value = value
                self.next = self
            def insertprev(self, node):
                self.prev.next = node
                node.prev = self.prev
                node.next = self
                self.prev = node
                return self
            def remove(self):
                if self.next is self:
                    return None
                self.prev.next = self.next
                self.next.prev = self.prev
                ret = self.next
                self.next = self
                self.prev = self
                return ret
        class CircleList(object):
            def __init__(self):
                self.current = None
            def remove(self, node):
                if self.current is node:
                    self.current = node.remove()
                else:
                    node.remove()
            def insertprev(self, node):
                if self.current is None:
                    self.current = node
                else:
                    self.current.insertprev(node)
            def insertcurrent(self, node):
                self.insertprev(node)
                self.current = node
            def next(self):
                ret = self.current
                if self.current is not None:
                    self.current = self.current.next
                return ret
            def clear(self):
                self.current = None
        def __init__(self, parent = None, priority = 0):
            self.queues = CBQueue.MultiQueue.CircleList()
            self.queueDict = {}
            self.queueStat = {}
            self.statseq = deque()
            self.parent = parent
            self.priority = priority
            self.totalSize = 0
            self.blocked = True
        def canPop(self):
            return bool(self.queues.current)
        def _pop(self):
            if not self.canPop():
                raise IndexError('pop from an empty or blocked queue')
            c = self.queues.next()
            ret = c.value._pop()
            self.queueStat[c.value] = self.queueStat.get(c.value, 0) + 1
            while len(self.statseq) >= 10 * len(self.queueDict) + 10:
                o = self.statseq.popleft()
                if o in self.queueStat:
                    self.queueStat[o] = self.queueStat[o] - 1
                    if self.queueStat[o] <= 0 and not o in self.queueDict:
                        del self.queueStat[o]
            self.statseq.append(c.value)
            if not c.value.canPop():
                self.queues.remove(c)
                self.queueDict[c.value] = None
            self.totalSize = self.totalSize - 1
            if not self.canPop():
                if not self.blocked:
                    self.blocked = True
                    if self.parent is not None:
                        self.parent.notifyBlock(self, True)
            return ret            
        def addSubQueue(self, queue):
            self.totalSize = self.totalSize + len(queue)
            queue.parent = self
            if queue.canPop():
                # Activate this queue
                node = CBQueue.MultiQueue.CircleListNode(queue)
                self.queues.insertprev(node)
                self.queueDict[queue] = node
                self.queueStat[queue] = 0
            else:
                self.queueDict[queue] = None
            if self.canPop():
                if self.blocked:
                    self.blocked = False
                    if self.parent is not None:
                        self.parent.notifyBlock(self, False)
        def removeSubQueue(self, queue):
            self.totalSize = self.totalSize - len(queue)
            if self.queueDict[queue] is not None:
                self.queues.remove(self.queueDict[queue])
            del self.queueDict[queue]
            if queue in self.queueStat:
                del self.queueStat[queue]
            if not self.canPop():
                if not self.blocked:
                    self.blocked = True
                    if self.parent is not None:
                        self.parent.notifyBlock(self, True)
        def notifyAppend(self, queue, force):
            if self.parent is not None:
                m = self.parent.notifyAppend(self, force)
                if m is not None:
                    return m
            self.totalSize = self.totalSize + 1
            if not queue.blocked:
                if self.queueDict[queue] is None:
                    # Activate this queue
                    node = CBQueue.MultiQueue.CircleListNode(queue)
                    qs = self.queueStat.setdefault(queue, 0)
                    if qs * len(self.queueStat) >= len(self.statseq):
                        self.queues.insertprev(node)
                    else:
                        self.queues.insertcurrent(node)
                    self.queueDict[queue] = node
            if self.canPop():
                if self.blocked:
                    self.blocked = False
                    if self.parent is not None:
                        self.parent.notifyBlock(self, False)
            return None
        def __len__(self):
            return self.totalSize
        def notifyBlock(self, queue, blocked):
            if queue.canPop():
                if self.queueDict[queue] is None:
                    # Activate this queue
                    node = CBQueue.MultiQueue.CircleListNode(queue)
                    qs = self.queueStat.setdefault(queue, 0)
                    if qs * len(self.queueStat) >= len(self.statseq):
                        self.queues.insertprev(node)
                    else:
                        self.queues.insertcurrent(node)
                    self.queueDict[queue] = node
            else:
                if self.queueDict[queue] is not None:
                    self.queues.remove(self.queueDict[queue])
                    self.queueDict[queue] = None
            selfblocked = not self.canPop()
            if selfblocked != self.blocked:
                self.blocked = selfblocked
                if self.parent is not None:
                    self.parent.notifyBlock(self, selfblocked)
        def notifyPop(self, queue, length = 1):
            self.totalSize = self.totalSize - length
            if not queue.canPop():
                if self.queueDict[queue] is not None:
                    self.queues.remove(self.queuDict[queue])
                    self.queueDict[queue] = None
            ret = ([], [])
            if self.parent is not None:
                ret = self.parent.notifyPop(self, length)
            blocked = not self.canPop()
            if blocked != self.blocked:
                self.blocked = blocked
                if self.parent is not None:
                    self.parent.notifyBlock(self, blocked)
            return ret
            
        def unblockall(self):
            for q in self.queueDict.keys():
                q.unblockall()
        def _clear(self):
            ret = ([],[])
            for q in self.queueDict.keys():
                pr = q._clear()
                ret[0].extend(pr[0])
                ret[1].extend(pr[1])
            self.totalSize = 0
            self.blockedSize = 0
            self.queues.clear()
            if not self.blocked:
                self.blocked = True
                if self.parent is not None:
                    self.parent.notifyBlock(self, True)
            return ret
    class AutoClassQueue(object):
        '''
        A queue classify events into virtual sub-queues by key
        '''
        nokey = object()
        def __init__(self, parent = None, maxlength = None, key = 'owner', preserveForNew = 1, maxstat = None, subqueuelimit = None):
            """
            Each value are classified and put into virtual subqueues.
            
            :param parent: parent queue
            
            :param maxlength: total limit for this queue
            
            :param key: classify value by this attribute
            
            :param preserveForNew: preserve some space for new subqueues. When putting values with old keys into
                                   the queue, the queue would block the value when it has size `maxlength - preserveForNew`.
                                   But, a value with new key (thus creates a new virtual queue) will be accepted by this
                                   queue. It makes sure a value with new key has enough priority to be enqueued.
            
            :param maxstat: Object passes each virtual queue lately are counted. When a virtual subqueue is passing
                            less events, the priority of this queue is increased to make it fair. `maxstat` is the limit
                            of objects counted
            
            :param subqueuelimit: limit of each subqueue
                            
            """
            self.queues = CBQueue.MultiQueue.CircleList()
            self.queueDict = {}
            self.queueStat = {}
            self.statseq = deque()
            self.maxlength = maxlength
            self.blocked = False
            if self.maxlength is not None and self.maxlength <= 0:
                self.maxlength = 1
            if maxstat is None:
                if self.maxlength is None:
                    self.maxstat = 10240
                else:
                    self.maxstat = maxlength * 10
            else:
                self.maxstat = maxstat
            if self.maxstat >= 10240:
                self.maxstat = 10240
            # Tuples with wanted notifications:
            # (False, False, key) - Any notification from subqueue
            # (False, True, key) - notification from subqueue, OR total limit with new-only = False, first-only = False
            # (False, False) - total limit with new-only = False, first-only = False
            # (False, True) - total limit with new-only = False, first-only = True/False
            # (True, True) - any total limit notifications
            self.waited = set()
            self.key = key
            self.preserve = preserveForNew
            self.totalSize = 0
            self.blockKeys = set()
            self.subqueuelimit = subqueuelimit
        @classmethod
        def initHelper(cls, key = 'owner', preserveForNew = 1, maxstat = None, subqueuelimit = None):
            def initer(parent = None, maxlength = None):
                return cls(parent, maxlength, key, preserveForNew, maxstat, subqueuelimit)
            return initer
        def append(self, value, force = False):
            key = getattr(value, self.key, self.nokey)
            # We use hash instead of reference or weakref, this may cause problem, but better than leak.
            kid = hash(key)
            if not force:
                w = self._tryAppend(key)
                if w is not None:
                    return w
            if self.parent is not None:
                m = self.parent.notifyAppend(self, force)
                if m is not None:
                    return m
            if key in self.queueDict:
                self.queueDict[key].value[1].append(value)
            else:
                node = CBQueue.MultiQueue.CircleListNode((key,deque()))
                node.value[1].append(value)
                qs = self.queueStat.setdefault(kid, 0)
                if qs * len(self.queueStat) >= len(self.statseq):
                    self.queues.insertprev(node)
                else:
                    self.queues.insertcurrent(node)
                self.queueDict[key] = node
            self.totalSize += 1
            blocked = not self.canPop() and self.totalSize > 0
            if blocked != self.blocked:
                self.blocked = blocked
                if self.parent is not None:
                    self.parent.notifyBlock(self, blocked)
            return None
        def _tryAppend(self, key):
            if self.maxlength is None:
                # No total length limit
                if self.subqueuelimit is None or not key in self.queueDict:
                    # No subqueue limit, or this is a new queue
                    return None
                elif len(self.queueDict[key].value[1]) >= self.subqueuelimit:
                    # Auto-subqueue limit exceeded, only appendable when this subqueue has more space
                    self.waited.add((False, False, key))
                    # Matches: any notification from this virtual queue
                    return AutoClassQueueCanWriteEvent.createMatcher(self, key)
                else:
                    # An old queue, but still has space
                    return None
            if key in self.queueDict:
                # Subqueue already exists
                if self.subqueuelimit is not None and len(self.queueDict[key].value[1]) >= self.subqueuelimit:
                    # Subqueue limit exceeded
                    # Wait for: any space from this virtual queue
                    self.waited.add((False, False, key))
                    # Matches: any notification from this virtual queue
                    return AutoClassQueueCanWriteEvent.createMatcher(self, key)
                elif self.totalSize < self.maxlength - self.preserve - len(self.queueStat) + len(self.queueDict):
                    # Total limit not exceeded
                    # An extra space is preserved for each empty virtual queue, so when there are more subqueues than
                    # the total limit, each queue can have at least one item
                    return None
                else:
                    # Blocked by total limit
                    if len(self.queueDict[key].value[1]) <= 1:
                        # When the only item in this queue is out, this queue becomes empty, and has a higher priority
                        # Wait for first-only notification
                        self.waited.add((False, True, key))
                        # Matches: newonly = False notification. If from other queue, firstonly = False.
                        return AutoClassQueueCanWriteEvent.createMatcher(self, None, False, _ismatch = lambda x: not (x.firstonly and x.key != key))
                    else:
                        # Wait for total limit change: any
                        self.waited.add((False, False))
                        # Matches: newonly = False, firstonly=False notifications from any virtual subqueue
                        return AutoClassQueueCanWriteEvent.createMatcher(self, None, False, False)
            elif hash(key) in self.queueStat:
                # An empty virtual queue
                if self.totalSize < self.maxlength - self.preserve:
                    return None
                else:
                    # Total limit exceeded
                    # Wait for total limit change: new-only = False
                    self.waited.add((False, True))
                    # Matches: notification from any virtual subqueue with new-only = False
                    return AutoClassQueueCanWriteEvent.createMatcher(self, None, False)
            else:
                # A new queue
                if self.totalSize < self.maxlength:
                    return None
                else:
                    # Total limit exceeded
                    # Wait for total limit change (any)
                    self.waited.add((True, True))
                    # Matches: any notification
                    return AutoClassQueueCanWriteEvent.createMatcher(self)
        def canAppend(self):
            return self.maxlength is None or self.totalSize < self.maxlength
        def canAppendSubqueue(self, key):
            if self.subqueuelimit is None or not key in self.queueDict:
                return True
            elif len(self.queueDict[key].value[1]) >= self.subqueuelimit:
                return False
            else:
                return True
        def canPop(self):
            return self.queues.current is not None
        def pop(self):
            ret = self._pop()
            if self.parent is not None:
                pr = self.parent.notifyPop(self)
                ret[1].extend(pr[0])
                ret[2].extend(pr[1])
            return ret
        def _pop(self):
            if not self.canPop():
                raise IndexError('pop from a blocked or empty queue')
            c = self.queues.next()
            key = c.value[0]
            kid = hash(key)
            ret = c.value[1].popleft()
            self.totalSize -= 1
            self.queueStat[kid] = self.queueStat.get(kid, 0) + 1
            while len(self.statseq) >= min(self.maxstat, 10 * len(self.queueStat) + 10):
                k1 = self.statseq.popleft()
                self.queueStat[k1] = self.queueStat[k1] - 1
                if self.queueStat[k1] <= 0:
                    del self.queueStat[k1]
            self.statseq.append(kid)
            if not c.value[1]:
                del self.queueDict[c.value[0]]
                self.queues.remove(c)
            blocked = not self.canPop() and self.totalSize > 0
            if blocked != self.blocked:
                self.blocked = blocked
                if self.parent is not None:
                    self.parent.notifyBlock(self, blocked)
            if self.waited:
                if key not in self.queueDict:
                    subsize = 0
                else:
                    subsize = len(self.queueDict[key].value[1])
                if self.maxlength is None:
                    if self.subqueuelimit is not None and subsize < self.subqueuelimit and (False, False, key) in self.waited:
                        # Subqueue has more space, wake up the specified queue
                        self.waited.discard((False, False, key))
                        return (ret, [AutoClassQueueCanWriteEvent(self, key, False, False)], [])
                elif self.totalSize < self.maxlength - self.preserve - len(self.queueStat) + len(self.queueDict):
                    # Whole queue has more space, wake up any waiter wait for the whole queue or this subqueue
                    self.waited = set(w for w in self.waited if len(w) == 3 and w[1] == False and w[2] != key)
                    return (ret, [AutoClassQueueCanWriteEvent(self, key, False, False)], [])
                elif self.totalSize < self.maxlength - self.preserve:
                    # Accept first item for each empty queue
                    if (False, True) in self.waited or (False, True, key) in self.waited or (True, True) in self.waited or \
                            (False, False, key) in self.waited:
                        self.waited.discard((False, True))
                        self.waited.discard((False, True, key))
                        self.waited.discard((True, True))
                        self.waited.discard((False, False, key))
                        # Do not wake up machers with firstonly=False
                        # Waiters waiting for the subqueue will be waken up, but they may get a new matcher for the total limit
                        return (ret, [AutoClassQueueCanWriteEvent(self, key, False, True)], [])
                elif self.totalSize < self.maxlength:
                    # Accept new queues only
                    if (True, True) in self.waited or (False, False, key) in self.waited or (False, True, key) in self.waited:
                        self.waited.discard((True, True))
                        self.waited.discard((False, False, key))
                        if (False, True, key) in self.waited:
                            # This virtual queue becomes empty, wake up the waiters and let them change the matcher
                            self.waited.discard((False, True, key))
                            return (ret, [AutoClassQueueCanWriteEvent(self, key, False, True)], [])
                        else:
                            return (ret, [AutoClassQueueCanWriteEvent(self, key, True, True)], [])
                elif self.subqueuelimit is not None and subsize < self.subqueuelimit and (False, False, key) in self.waited:
                    # Wake up waiters waiting for this sub queue, let them change the matcher
                    
                    # If we don't wake up the sub-queue waiter now, it may wait forever.
                    # The sub-queue waiter won't be able to send events in, but they will get a new matcher
                    # Some waiters might wake up mistakenly, they will wait again when they try to append the event. 
                    self.waited.discard((True, True))
                    self.waited.discard((False, False, key))
                    return (ret, [AutoClassQueueCanWriteEvent(self, key, True, True)], [])
            return (ret, [], [])
        def clear(self):
            l = len(self)
            ret = self._clear()
            if self.parent is not None:
                pr = self.parent.notifyPop(self, l)
                ret[0].extend(pr[0])
                ret[1].extend(pr[1])
            return ret
        def _clear(self):
            self.queues.clear()
            self.blockKeys.clear()
            self.queueDict.clear()
            self.totalSize = 0
            blocked = not self.canPop() and self.totalSize > 0
            if blocked != self.blocked:
                self.blocked = blocked
                if self.parent is not None:
                    self.parent.notifyBlock(self, blocked)
            if self.waited:
                key_events = [AutoClassQueueCanWriteEvent(self, w[2], False, False)
                              for w in self.waited
                              if len(w) == 3]
                if not key_events:
                    key_events.append(AutoClassQueueCanWriteEvent(self, self.nokey, False, False))
                self.waited.clear()
                return (key_events, [])
            else:
                return ([], [])
        def __len__(self):
            return self.totalSize
        def block(self, value):
            if self.parent is not None:
                self.parent.notifyAppend(self, True)
            key = getattr(value, self.key, self.nokey)
            if key in self.queueDict:
                self.queueDict[key].value[1].appendleft(value)
                self.queues.remove(self.queueDict[key])
            else:
                node = CBQueue.MultiQueue.CircleListNode((key,deque()))
                node.value[1].append(value)
                self.queueDict[key] = node
            self.blockKeys.add(key)
            self.totalSize += 1
            blocked = not self.canPop() and self.totalSize > 0
            if blocked != self.blocked:
                self.blocked = blocked
                if self.parent is not None:
                    self.parent.notifyBlock(self, blocked)
        def unblock(self, value):
            key = getattr(value, self.key, self.nokey)
            if key in self.blockKeys:
                self._unblock(key)
                blocked = not self.canPop() and self.totalSize > 0
                if blocked != self.blocked:
                    self.blocked = blocked
                    if self.parent is not None:
                        self.parent.notifyBlock(self, blocked)
        def _unblock(self, key):
            self.blockKeys.remove(key)
            node = self.queueDict[key]
            qs = self.queueStat.setdefault(hash(key), 0)
            if qs * len(self.queueStat) >= len(self.statseq):
                self.queues.insertprev(node)
            else:
                self.queues.insertcurrent(node)            
        def unblockall(self):
            for k in list(self.blockKeys):
                self._unblock(k)
            blocked = not self.canPop() and self.totalSize > 0
            if blocked != self.blocked:
                self.blocked = blocked
                if self.parent is not None:
                    self.parent.notifyBlock(self, blocked)
    def __init__(self, tree = None, parent = None, maxdefault = None, maxtotal = None, defaultQueueClass = FifoQueue, defaultQueuePriority = 0):
        '''
        Constructor
        '''
        self.queues = {}
        self.queueindex = {}
        self.prioritySet = []
        if tree is None:
            self.tree = MatchTree()
        else:
            self.tree = tree
        self.parent = parent
        defaultPriority = CBQueue.MultiQueue(self, defaultQueuePriority)
        defaultQueue = defaultQueueClass(defaultPriority, maxdefault)
        defaultPriority.addSubQueue(defaultQueue)
        self.queues[defaultQueuePriority] = defaultPriority
        self.tree.insert(None, defaultQueue)
        self.defaultQueue = defaultQueue
        self.totalSize = 0
        self.maxtotal = maxtotal
        self.blocked = True
        self.blockEvents = {}
        self.isWaited = False
        self.isWaitEmpty = False
        self.outputStat = 0
        self._matcher = QueueCanWriteEvent.createMatcher(self)
    def _removeFromTree(self):
        for v in self.queueindex.values():
            if len(v) == 3:
                v[1]._removeFromTree()
        self.tree.remove(None, self.defaultQueue)
        self.tree = None
    def canAppend(self):
        '''
        Whether the queue is full or not. Only check the total limit. Sub-queue may still be full (even default).
        
        :returns: False if the queue is full, True if not.
                  If there are sub-queues, append() may still fail if the sub-queue is full. 
        '''
        return self.maxtotal is None or self.totalSize < self.maxtotal
    def append(self, event, force = False):
        '''
        Append an event to queue. The events are classified and appended to sub-queues
        
        :param event: input event
        
        :param force: if True, the event is appended even if the queue is full
        
        :returns: None if appended successfully, or a matcher to match a QueueCanWriteEvent otherwise
        '''
        if self.tree is None:
            if self.parent is None:
                raise IndexError('The queue is removed')
            else:
                return self.parent.parent.append(event, force)
        q = self.tree.matchfirst(event)
        return q.append(event, force)
    def waitForEmpty(self):
        '''
        Make this queue generate a QueueIsEmptyEvent when it is empty
        
        :returns: matcher for QueueIsEmptyEvent, or None if the queue is already empty
        '''
        if not self:
            return None
        self.isWaitEmpty = True
        return QueueIsEmptyEvent.createMatcher(self)
    def block(self, event, emptyEvents = ()):
        '''
        Return a recently popped event to queue, and block all later events until unblock.
        
        Only the sub-queue directly containing the event is blocked, so events in other queues may still be processed.
        It is illegal to call block and unblock in different queues with a same event.
        
        :param event: the returned event. When the queue is unblocked later, this event will be popped again.
        
        :param emptyEvents: reactivate the QueueIsEmptyEvents
        '''
        q = self.tree.matchfirst(event)
        q.block(event)
        self.blockEvents[event] = q
        for ee in emptyEvents:
            ee.queue.waitForEmpty()
    def unblock(self, event):
        '''
        Remove a block 
        '''
        if event not in self.blockEvents:
            return
        self.blockEvents[event].unblock(event)
        del self.blockEvents[event]
    def unblockqueue(self, queue):
        '''
        Remove blocked events from the queue and all subqueues. Usually used after queue clear/unblockall to prevent leak.
        
        :returns: the cleared events
        '''
        subqueues = set()
        def allSubqueues(q):
            subqueues.add(q)
            subqueues.add(q.defaultQueue)
            for v in q.queueindex.values():
                if len(v) == 3:
                    allSubqueues(v[1])
        allSubqueues(queue)
        events = [k for k,v in self.blockEvents.items() if v in subqueues]
        for e in events:
            del self.blockEvents[e]
        return events
    def unblockall(self):
        '''
        Remove all blocks from the queue and all sub-queues
        '''
        for q in self.queues.values():
            q.unblockall()
        self.blockEvents.clear()
    def notifyAppend(self, queue, force):
        '''
        Internal notify for sub-queues
        
        :returns: If the append is blocked by parent, an EventMatcher is returned, None else.
        '''
        if not force and not self.canAppend():
            self.isWaited = True
            return self._matcher
        if self.parent is not None:
            m = self.parent.notifyAppend(self, force)
            if m is not None:
                return m
        self.totalSize = self.totalSize + 1
        return None
    def notifyBlock(self, queue, blocked):
        '''
        Internal notify for sub-queues been blocked
        '''
        if blocked:
            if self.prioritySet[-1] == queue.priority:
                self.prioritySet.pop()
            else:
                pindex = bisect_left(self.prioritySet, queue.priority)
                if pindex < len(self.prioritySet) and self.prioritySet[pindex] == queue.priority:
                    del self.prioritySet[pindex]
        else:
            if queue.canPop():
                pindex = bisect_left(self.prioritySet, queue.priority)
                if pindex >= len(self.prioritySet) or self.prioritySet[pindex] != queue.priority:
                    self.prioritySet.insert(pindex, queue.priority)
        newblocked =  not self.canPop()
        if newblocked != self.blocked:
            self.blocked = newblocked
            if self.parent is not None:
                self.parent.notifyBlock(self, newblocked)
    def notifyPop(self, queue, length = 1):
        '''
        Internal notify for sub-queues been poped
        
        :returns: List of any events generated by this pop
        '''
        self.totalSize = self.totalSize - length
        ret1 = []
        ret2 = []
        if self.isWaited and self.canAppend():
            self.isWaited = False
            ret1.append(QueueCanWriteEvent(self))
        if self.isWaitEmpty and not self:
            self.isWaitEmpty = False
            ret2.append(QueueIsEmptyEvent(self))
        if self.parent is not None:
            pr = self.parent.notifyPop(self, length)
            ret1 += pr[0]
            ret2 += pr[1]
        newblocked =  not self.canPop()
        if newblocked != self.blocked:
            self.blocked = newblocked
            if self.parent is not None:
                self.parent.notifyBlock(self, newblocked)
        return (ret1, ret2)
    def canPop(self):
        '''
        Whether the queue is empty/blocked or not
        
        :returns: False if the queue is empty or blocked, or True otherwise
        '''
        return bool(self.prioritySet)
    def pop(self):
        '''
        Pop an event from the queue. The event in the queue with higher priority is popped before ones in lower priority.
        If there are multiple queues with the same priority, events are taken in turn from each queue.
        May return some queueEvents indicating that some of the queues can be written into.
        
        :returns: `(obj, (queueEvents,...), (queueEmptyEvents,...))` where obj is the popped event, queueEvents are QueueCanWriteEvents generated by this pop
                  and queueEmptyEvents are QueueIsEmptyEvents generated by this pop
        '''
        ret = self._pop()
        if self.parent is not None:
            pr = self.parent.notifyPop(self)
            ret[1].extend(pr[0])
            ret[2].extend(pr[1])
        return ret
    def _pop(self):
        '''
        Actual pop
        '''
        if not self.canPop():
            raise IndexError('pop from an empty or blocked queue')
        priority = self.prioritySet[-1]
        ret = self.queues[priority]._pop()
        self.outputStat = self.outputStat + 1
        self.totalSize = self.totalSize - 1
        if self.isWaited and self.canAppend():
            self.isWaited = False
            ret[1].append(QueueCanWriteEvent(self))
        if self.isWaitEmpty and not self:
            self.isWaitEmpty = False
            ret[2].append(QueueIsEmptyEvent(self))
        return ret
    def clear(self):
        '''
        Clear all the events in this queue, including any sub-queues.
        
        :returns: ((queueEvents,...), (queueEmptyEvents,...)) where queueEvents are QueueCanWriteEvents generated by clearing.
        '''
        l = len(self)
        ret = self._clear()
        if self.parent is not None:
            pr = self.parent.notifyPop(self, l)
            ret[0].extend(pr[0])
            ret[1].extend(pr[1])
        return ret
    def _clear(self):
        '''
        Actual clear
        '''
        ret = ([],[])
        for q in self.queues.values():
            pr = q._clear()
            ret[0].extend(pr[0])
            ret[1].extend(pr[1])
        self.totalSize = 0
        del self.prioritySet[:]
        if self.isWaited and self.canAppend():
            self.isWaited = False
            ret[0].append(QueueCanWriteEvent(self))
        if self.isWaitEmpty and not self:
            self.isWaitEmpty = False
            ret[1].append(QueueIsEmptyEvent(self))
        self.blockEvents.clear()
        return ret
    def __contains__(self, name):
        return name in self.queueindex
    def __getitem__(self, name):
        '''
        Get a sub-queue through q['sub-queue-name']
        '''
        return self.queueindex[name][1]
    def getPriority(self, queue):
        '''
        get priority of a sub-queue
        '''
        return self.queueindex[queue][0]
    def setPriority(self, queue, priority):
        '''
        Set priority of a sub-queue
        '''
        q = self.queueindex[queue]
        self.queues[q[0]].removeSubQueue(q[1])
        newPriority = self.queues.setdefault(priority, CBQueue.MultiQueue(self, priority))
        q[0] = priority
        newPriority.addSubQueue(q[1])
        
    def addSubQueue(self, priority, matcher, name = None, maxdefault = None, maxtotal = None, defaultQueueClass = FifoQueue):
        '''
        add a sub queue to current queue, with a priority and a matcher
        
        :param priority: priority of this queue. Larger is higher, 0 is lowest.
        
        :param matcher: an event matcher to catch events. Every event match the criteria will be stored in this queue.
        
        :param name: a unique name to identify the sub-queue. If none, the queue is anonymous. It can be any hashable value.
        
        :param maxdefault: max length for default queue.
        
        :param maxtotal: max length for sub-queue total, including sub-queues of sub-queue
        '''
        if name is not None and name in self.queueindex:
            raise IndexError("Duplicated sub-queue name '" + str(name) + "'")
        subtree = self.tree.subtree(matcher, True)
        newPriority = self.queues.setdefault(priority, CBQueue.MultiQueue(self, priority))
        newQueue = CBQueue(subtree, newPriority, maxdefault, maxtotal, defaultQueueClass)
        newPriority.addSubQueue(newQueue)
        qi = [priority, newQueue, name]
        if name is not None:
            self.queueindex[name] = qi
        self.queueindex[newQueue] = qi
        return newQueue
    def removeSubQueue(self, queue):
        '''
        remove a sub queue from current queue.
        
        This unblock the sub-queue, retrieve all events from the queue and put them back to the parent.
        
        Call clear on the sub-queue first if the events are not needed any more.
        
        :param queue: the name or queue object to remove
        
        :returns: ((queueevents,...), (queueEmptyEvents,...)) Possible queue events from removing sub-queues
        '''
        q = self.queueindex[queue]
        q[1].unblockall()
        q[1]._removeFromTree()
        ret = ([],[])
        while q[1].canPop():
            r = q[1].pop()
            self.append(r[0], True)
            ret[0].extend(r[1])
            ret[1].extend(r[2])
        self.queues[q[0]].removeSubQueue(q[1])
        # Remove from index
        if q[2] is not None:
            del self.queueindex[q[2]]
        del self.queueindex[q[1]]
        newblocked =  not self.canPop()
        if newblocked != self.blocked:
            self.blocked = newblocked
            if self.parent is not None:
                self.parent.notifyBlock(self, newblocked)
        return ret
    def __len__(self):
        return self.totalSize
        



from vlcp.config import defaultconfig
from vlcp.event import Event
from vlcp.event import RoutineContainer
from vlcp.event import withIndices
from vlcp.protocol.openflow import OpenflowConnectionStateEvent
from vlcp.server.module import call_api, publicapi, api
from vlcp.service.sdn import ioprocessing
from vlcp.service.sdn import ovsdbportmanager
from vlcp.service.sdn.flowbase import FlowBase
from vlcp.service.sdn.ofpmanager import FlowInitialize
from vlcp.service.utils.remoteapi import remoteAPI
from vlcp.utils.ethernet import ETHERTYPE_8021Q
from contextlib import closing
from vlcp.event.event import M_


def _is_vxlan(obj):
    try:
        return obj.physicalnetwork.type == 'vxlan'
    except AttributeError:
        return False


@withIndices("connection","logicalnetworkid","type")
class VXLANMapChanged(Event):
    UPDATED = 'updated'
    DELETED = 'deleted'

@withIndices("connection","physname","phyiname","type")
class VtepControllerCall(Event):
    BIND = 'bind'
    UNBIND = "unbind"
    UNBINDALL = "unbindall"

class VXLANHandler(RoutineContainer):
    def __init__(self,connection,parent):
        super(VXLANHandler,self).__init__(connection.scheduler)
        self._conn = connection
        self._parent = parent

        self.vxlan_vlan_map_info = {}
        self.vlanid_pool = parent.vlanid_pool
        self._last_lognets_info = {}
        self._last_phyport_info = {}

        self._store_event = dict()
        self._store_bind_event = []
        self._store_unbind_event = []
        self._store_unbindall_event = []

        if self._parent.remote_api:
            self.api = remoteAPI
        else:
            self.api = call_api

    async def main(self):

        try:
            data_change_event = ioprocessing.DataObjectChanged.createMatcher(None, None, self._conn)

            # routine sync call to vtep controller (updatelogicalswitch,unbindlogicalswitch,unbindphysicalport)
            self.subroutine(self.action_handler(),True,"_action_handler")

            _last_change_event = None
            while True:
                ev = await data_change_event
                _last_change_event = ev
                def _update_callback(event, matcher):
                    nonlocal _last_change_event
                    _last_change_event = event
                async def _update_loop():
                    nonlocal _last_change_event
                    while _last_change_event is not None:
                        try:
                            last_lognets_info = self._last_lognets_info
                            last_phyport_info = self._last_phyport_info
        
                            current_logports,current_phyports,current_lognets,_ = _last_change_event.current
                            _last_change_event = None
        
                            lognet_to_logports = {}
                            for p, _ in current_logports:
                                lognet_to_logports.setdefault(p.network, []).append(p.id)
        
                            datapath_id = self._conn.openflow_datapathid
                            vhost = self._conn.protocol.vhost
        
                            async def get_phyport_info(portid):
                                try:
                                    return await call_api(self, "ovsdbportmanager", "waitportbyno",
                                                 {"datapathid": datapath_id,
                                                  "vhost": vhost,
                                                  "timeout": 1,
                                                  "portno": portid})
                                except ovsdbportmanager.OVSDBPortNotAppearException:
                                    return None
                            phy_ports = [p for p in current_phyports if _is_vxlan(p[0])]
                            port_result = await self.execute_all([get_phyport_info(pid) for p,pid in phy_ports])
        
                            phy_port_info  = dict((k[0], (v['external_ids']['vtep-physname'],
                                                                v['external_ids']['vtep-phyiname']))
                                                  for k,v in zip(phy_ports, port_result)
                                                  if v is not None and \
                                                    'vtep-physname' in v['external_ids'] and \
                                                    'vtep-phyiname' in v['external_ids'])
                            unique_phyports = dict((p.physicalnetwork, p)
                                                   for p,_ in sorted(phy_ports, key = lambda x: x[1])
                                                   if p in phy_port_info)
        
                            lognet_to_phyport =  dict((n,unique_phyports[n.physicalnetwork])
                                                      for n,_ in current_lognets
                                                      if n.physicalnetwork in unique_phyports)
        
                            current_lognets_info = dict((n,(phy_port_info[v], lognet_to_logports[n]))
                                                        for n,v in lognet_to_phyport.items()
                                                        if n in lognet_to_logports)
        
                            self._last_lognets_info = current_lognets_info
                            self._last_phyport_info = phy_port_info
        
                            # add or update new phyport, unbind all physical interface bind
                            # must do it as init first ....
                            for p in phy_port_info:
                                if p not in last_phyport_info or\
                                        (p in last_phyport_info and last_phyport_info[p] != phy_port_info[p]):
        
                                    physname = phy_port_info[p][0]
                                    phyiname = phy_port_info[p][1]
        
                                    event = VtepControllerCall(connection=self._conn,type=VtepControllerCall.UNBINDALL,
                                                               physname=physname,phyiname=phyiname)
        
                                    await self.wait_for_send(event)
        
        
                            # remove phyport , unbind all physical interface bind also
                            for p in last_phyport_info:
                                if p not in phy_port_info or\
                                        (p in phy_port_info and last_phyport_info[p] != phy_port_info[p]):
                                    physname = last_phyport_info[p][0]
                                    phyiname = last_phyport_info[p][1]
        
                                    event = VtepControllerCall(connection=self._conn,type=VtepControllerCall.UNBINDALL,
                                                               physname=physname,phyiname=phyiname)
        
                                    await self.wait_for_send(event)
        
                            for n in last_lognets_info:
                                if n not in current_lognets_info:
                                    # means lognet removed
                                    # release vlan id
                                    # unbind physwitch
        
                                    # if last cause exception , n.id mybe not in vxlan_vlan_map_info
                                    if n.id in self.vxlan_vlan_map_info:
                                        vlan_id = self.vxlan_vlan_map_info[n.id]
                                        physname = last_lognets_info[n][0][0]
                                        phyiname = last_lognets_info[n][0][1]
        
                                        del self.vxlan_vlan_map_info[n.id]
        
                                        event = VXLANMapChanged(self._conn,n.id,VXLANMapChanged.DELETED)
        
                                        await self.wait_for_send(event)
        
                                        event = VtepControllerCall(connection=self._conn,logicalnetworkid=n.id,
                                                                   physname=physname,phyiname=phyiname,vlanid=vlan_id,
                                                                   type=VtepControllerCall.UNBIND)
        
                                        await self.wait_for_send(event)
                                
                                else:
                                    last_info = last_lognets_info[n]
                                    current_info = current_lognets_info[n]
        
                                    if last_info[0] != current_info[0]:
                                        # physwitch info have changed
                                        # unbind it
        
                                        if n.id in self.vxlan_vlan_map_info:
                                            vid = self.vxlan_vlan_map_info[n.id]
                                            physname = last_lognets_info[n][0][0]
                                            phyiname = last_lognets_info[n][0][1]
        
                                            del self.vxlan_vlan_map_info[n.id]
        
                                            event = VtepControllerCall(connection=self._conn, logicalnetworkid=n.id,
                                                                       physname=physname, phyiname=phyiname, vlanid=vid,
                                                                       type=VtepControllerCall.UNBIND)
        
                                            await self.wait_for_send(event)
                                           
                                    else:
                                        if last_info[1] != current_info[1]:        
                                            if set(current_info[1]) != set(last_info[1]):
                                                # update bind logical ports
        
                                                if n.id in self.vxlan_vlan_map_info:
                                                    vid = self.vxlan_vlan_map_info[n.id]
                                                    physname = last_lognets_info[n][0][0]
                                                    phyiname = last_lognets_info[n][0][1]
        
                                                    event = VtepControllerCall(connection=self._conn, logicalnetworkid=n.id,
                                                                               vni = n.vni,logicalports=current_info[1],
                                                                               physname=physname, phyiname=phyiname, vlanid=vid,
                                                                               type=VtepControllerCall.BIND)
        
                                                    await self.wait_for_send(event)
        
        
                            for n in current_lognets_info:
                                if n not in last_lognets_info:
                                    # means add lognet
                                    # find avaliable vlan id , send event
                                    # bind physwitch
                                    find = False
                                    _current_vlan_ids = set(self.vxlan_vlan_map_info.values())
                                    for start,end in self.vlanid_pool:
                                        for vid in range(start,end + 1):
                                            if vid not in _current_vlan_ids:
                                                find = True
                                                break
                                        if find:
                                            break
        
                                    if find:
                                        self.vxlan_vlan_map_info[n.id] = vid
        
                                        event = VXLANMapChanged(self._conn,n.id,VXLANMapChanged.UPDATED,vlan_id = vid)
                                        await self.wait_for_send(event)
        
                                        physname = current_lognets_info[n][0][0]
                                        phyiname = current_lognets_info[n][0][1]
                                        ports = current_lognets_info[n][1]
        
                                        event = VtepControllerCall(connection=self._conn, logicalnetworkid=n.id,
                                                                   vni=n.vni, logicalports=ports,
                                                                   physname=physname, phyiname=phyiname, vlanid=vid,
                                                                   type=VtepControllerCall.BIND)
        
                                        await self.wait_for_send(event)
        
                                    else:
                                        self._parent._logger.warning('Not enough vlan_id for logical network %r', n.id)
                                        event = VXLANMapChanged(self._conn,n.id,VXLANMapChanged.UPDATED,vlan_id = None)
                                        await self.wait_for_send(event)
                                else:
        
                                    last_info = last_lognets_info[n]
                                    current_info = current_lognets_info[n]
        
                                    if last_info[0] != current_info[0]:
                                        # physwitch info have changed
                                        # rebind it
                                        if n.id in self.vxlan_vlan_map_info:
                                            vid = self.vxlan_vlan_map_info[n.id]
                                            physname = current_lognets_info[n][0][0]
                                            phyiname = current_lognets_info[n][0][1]
                                            ports = current_lognets_info[n][1]
        
                                            event = VtepControllerCall(connection=self._conn, logicalnetworkid=n.id,
                                                                       vni=n.vni, logicalports=ports,
                                                                       physname=physname, phyiname=phyiname, vlanid=vid,
                                                                       type=VtepControllerCall.BIND)
        
                                            await self.wait_for_send(event)
        
        
                        except Exception:
                            self._parent._logger.info(" vxlan vtep handler exception , continue", exc_info = True)
                await self.with_callback(_update_loop(), _update_callback, data_change_event)
        finally:

            if hasattr(self,"_action_handler"):
                self._action_handler.close()


# ===============================================================================
#     def time_cycle(self):
#
#         while True:
#             for m in self.waitWithTimeout(self._parent.refreshinterval):
#                 yield m
#
#             for n,v in self._last_lognets_info.items():
#
#                 if n.id in self.vxlan_vlan_map_info:
#                     vid = self.vxlan_vlan_map_info[n.id]
#                     physname = v[0][0]
#                     phyiname = v[0][1]
#                     ports = v[1]
#
#                     event = VtepControllerCall(connection=self._conn, logicalnetworkid=n.id,
#                                                                vni=n.vni, logicalports=ports,
#                                                                physname=physname, phyiname=phyiname, vlanid=vid,
#                                                                type=VtepControllerCall.BIND)
#                     self.scheduler.emergesend(event)
# ===============================================================================

    async def action_handler(self):
        """
        Call vtep controller in sequence, merge mutiple calls if possible
        
        When a bind relationship is updated, we always send all logical ports to a logicalswitch,
        to make sure it recovers from some failed updates (so called idempotency). When multiple
        calls are pending, we only need to send the last of them.
        """
        bind_event = VtepControllerCall.createMatcher(self._conn)
        event_queue = []
        timeout_flag = [False]

        async def handle_action():
            while event_queue or timeout_flag[0]:
                events = event_queue[:]
                del event_queue[:]

                for e in events:
                    # every event must have physname , phyiname
                    # physname: physical switch name - must be same with OVSDB-VTEP switch
                    # phyiname: physical port name - must be same with the corresponding port
                    physname = e.physname
                    phyiname = e.phyiname
                    if e.type == VtepControllerCall.UNBINDALL:
                        # clear all other event info
                        self._store_event[(physname,phyiname)] = {"all":e}
                    elif e.type == VtepControllerCall.BIND:
                        # bind will combine bind event before
                        vlanid = e.vlanid
                        if (physname,phyiname) in self._store_event:
                            v = self._store_event[(physname,phyiname)]

                            if vlanid in v:
                                logicalports = e.logicalports
                                v.update({vlanid:(e.type,e.logicalnetworkid,e.vni,logicalports)})
                                self._store_event[(physname,phyiname)] = v
                            else:
                                # new bind info , no combind event
                                v.update({vlanid:(e.type,e.logicalnetworkid,e.vni,e.logicalports)})
                                self._store_event[(physname,phyiname)] = v
                        else:
                            self._store_event[(physname,phyiname)] = {vlanid:(e.type,e.logicalnetworkid,
                                                                              e.vni,e.logicalports)}

                    elif e.type == VtepControllerCall.UNBIND:

                        vlanid = e.vlanid

                        if (physname,phyiname) in self._store_event:
                            v = self._store_event[(physname,phyiname)]
                            v.update({vlanid:(e.type,e.logicalnetworkid)})
                            self._store_event[(physname,phyiname)] = v
                        else:
                            self._store_event[(physname,phyiname)] = {vlanid:(e.type,e.logicalnetworkid)}

                    else:
                        self._parent._logger.warning("catch error type event %r , ignore it", exc_info=True)
                        continue

                call = []
                target_name = "vtepcontroller"
                for k,v in self._store_event.items():
                    if "all" in v:
                        # send unbindall
                        call.append(self.api(self,target_name,"unbindphysicalport",
                                             {"physicalswitch": k[0], "physicalport": k[1]},
                                             timeout=10))
                        # unbindall , del it whatever
                        del v["all"]

                try:
                    await self.execute_all(call)
                except Exception:
                    self._parent._logger.warning("unbindall remove call failed", exc_info=True)

                for k,v in self._store_event.items():
                    for vlanid , e in dict(v).items():
                        if vlanid != "all":
                            if e[0] == VtepControllerCall.BIND:

                                params = {"physicalswitch": k[0],
                                            "physicalport": k[1],
                                            "vlanid": vlanid,
                                            "logicalnetwork": e[1],
                                            "vni":e[2],
                                            "logicalports": e[3]}

                                try:
                                    await self.api(self,target_name,"updatelogicalswitch",
                                                  params,timeout=10)
                                except Exception:
                                    self._parent._logger.warning("update logical switch error,try next %r",params, exc_info=True)
                                else:
                                    del self._store_event[k][vlanid]

                            elif e[0] == VtepControllerCall.UNBIND:

                                params = {"logicalnetwork":e[1],
                                                "physicalswitch":k[0],
                                                "physicalport":k[1],
                                                  "vlanid":vlanid}

                                try:
                                    await self.api(self,target_name,"unbindlogicalswitch",
                                                      params,timeout=10)
                                except Exception:
                                    self._parent._logger.warning("unbind logical switch error,try next %r",params, exc_info=True)
                                else:
                                    del self._store_event[k][vlanid]

                self._store_event = dict((k,v) for k,v in self._store_event.items() if v)

                if timeout_flag[0]:
                    timeout_flag[0] = False

        def append_event(event, matcher):
            event_queue.append(event)

        while True:
            timeout, ev, m = await self.wait_with_timeout(10, bind_event)

            if not timeout:
                event_queue.append(ev)
            else:
                timeout_flag[0] = True

            await self.with_callback(handle_action(), append_event, bind_event)


    async def wait_vxlan_map_info(self,container,logicalnetworkid,timeout = 4):

        if logicalnetworkid in self.vxlan_vlan_map_info:
            return self.vxlan_vlan_map_info[logicalnetworkid]
        else:
            matcher = VXLANMapChanged.createMatcher(self._conn,logicalnetworkid,VXLANMapChanged.UPDATED)
            timeout_, ev, m = await container.wait_with_timeout(timeout, matcher)

            if timeout_:
                raise ValueError(" cannot find vxlan vlan map info ")
            else:
                return ev.vlan_id


def _check_vlanrange(vlanrange):
    lastend = 0
    for start,end in vlanrange:
        if start <= 0 or end > 4095:
            raise ValueError('VLAN ID out of range (1 - 4095)')
        if start > end or start <= lastend:
            raise ValueError('VLAN sequences overlapped or disordered: [%r, %r]' % (start, end))
        lastend = end


@defaultconfig
class VXLANVtep(FlowBase):
    """
    Use hardware_vtep instead of software VXLAN
    """
    # Use these VLANs for vtep configuration. Must not be conflicted with VLAN networks.
    _default_vlanid_pool = ((3000, 4000),)

    # remote_api means call remote api to vtep controller
    _default_remote_api = True

    # interval time to retry failed actions
    _default_retryactioninterval = 60

    def __init__(self,server):
        super(VXLANVtep,self).__init__(server)
        self.app_routine = RoutineContainer(self.scheduler)
        self.app_routine.main = self._main
        self.routines.append(self.app_routine)

        self.conns = {}
        self.vxlan_vlan_map_info = {}

        self.createAPI(publicapi(self.createioflowparts, self.app_routine,
                                 lambda connection,logicalnetwork,**kwargs:
                                        _is_vxlan(logicalnetwork)),
                       api(self.get_vxlan_bind_info,self.app_routine))

    async def _main(self):

        # check vlan pool
        lastend = 0
        _check_vlanrange(self.vlanid_pool)

        flow_init = FlowInitialize.createMatcher(_ismatch= lambda x : self.vhostbind is None
                                                 or x.vhost in self.vhostbind)
        conn_down = OpenflowConnectionStateEvent.createMatcher(state = OpenflowConnectionStateEvent.CONNECTION_DOWN,
                                                               _ismatch= lambda x:self.vhostbind is None
                                                               or x.createby.vhost in self.vhostbind)

        while True:
            ev, m = await M_(flow_init,conn_down)

            if m is flow_init:

                conn = ev.connection

                self.app_routine.subroutine(self._init_conn(conn))

            elif m is conn_down:
                conn = ev.connection

                self.app_routine.subroutine(self._uninit_conn(conn))

    async def _init_conn(self,conn):
        if conn in self.conns:
            handler = self.conns.pop(conn)
            handler.close()

        handler = VXLANHandler(conn,self)
        handler.start()
        self.conns[conn] = handler

    async def _uninit_conn(self,conn):
        if conn in self.conns:
            handler = self.conns.pop(conn)
            handler.close()

    async def get_vxlan_bind_info(self,systemid=None):
        """get vxlan -> vlan , bind info"""

        ret = []
        for conn in self.conns:
            datapath_id = conn.openflow_datapathid
            vhost = conn.protocol.vhost

            r = await call_api(self.app_routine,"ovsdbmanager","getbridgeinfo",{"datapathid":datapath_id,
                                                                              "vhost":vhost})

            if r:
                _,system_id,_ = r

                if systemid:
                    if systemid == system_id:
                        handler = self.conns[conn]
                        ret.append({system_id: handler.vxlan_vlan_map_info})
                        break
                else:
                    handler = self.conns[conn]
                    ret.append({system_id:handler.vxlan_vlan_map_info})


        return ret

    async def createioflowparts(self,connection,logicalnetwork,physicalport,logicalnetworkid,physicalportid):

        self._logger.debug(" create flow parts connection = %r",connection)
        self._logger.debug(" create flow parts logicalnetworkid = %r", logicalnetworkid)
        self._logger.debug(" create flow parts physicalportid = %r", physicalportid)

        find = False
        vid = None
        if connection in self.conns:

            handler = self.conns[connection]

            try:
                vid = await handler.wait_vxlan_map_info(self.app_routine,logicalnetwork.id)
            except Exception:
                find = False
                self._logger.warning(" get vxlan vlan map info error", exc_info = True)
            else:
                find = True
                if vid is None:
                    self._logger.warning('Not enough vlan ID, io flow parts not created')
                    find = False
        if find:
            return self.createflowparts(connection, vid, physicalportid)
        else:
            return ([],[],[],[],[])

    def createflowparts(self,connection,vid,physicalportid):
        ofdef = connection.openflowdef
        in_flow_match_part = [
            ofdef.create_oxm(ofdef.OXM_OF_VLAN_VID, vid | ofdef.OFPVID_PRESENT)
        ]

        in_flow_action_part = [
            ofdef.ofp_action(type=ofdef.OFPAT_POP_VLAN)
        ]

        out_flow_action_part = [
            ofdef.ofp_action_push(ethertype=ETHERTYPE_8021Q),
            ofdef.ofp_action_set_field(
                field=ofdef.create_oxm(
                    ofdef.OXM_OF_VLAN_VID,
                    vid | ofdef.OFPVID_PRESENT
                )
            ),
            ofdef.ofp_action_output(
                port=physicalportid
            )
        ]

        out_flow_action_part2 = [
            ofdef.ofp_action_push(ethertype=ETHERTYPE_8021Q),
            ofdef.ofp_action_set_field(
                field=ofdef.create_oxm(
                    ofdef.OXM_OF_VLAN_VID,
                    vid | ofdef.OFPVID_PRESENT
                )
            ),
            ofdef.ofp_action_output(
                port=ofdef.OFPP_IN_PORT
            )
        ]

        group_buckets = [
            ofdef.ofp_action_push(ethertype=ETHERTYPE_8021Q),
            ofdef.ofp_action_set_field(
                field=ofdef.create_oxm(
                    ofdef.OXM_OF_VLAN_VID,
                    vid | ofdef.OFPVID_PRESENT
                )
            ),
            ofdef.ofp_action_output(
                port=physicalportid
            )
        ]

        return in_flow_match_part, in_flow_action_part,out_flow_action_part, group_buckets,out_flow_action_part2



'''
Created on 2015/7/30

:author: hubo
'''
from . import openflow10
from . import openflow13
from . import openflow14
definations = {openflow10.OFP_VERSION: openflow10,
               openflow13.OFP_VERSION: openflow13,
               openflow14.OFP_VERSION: openflow14}



'''
Created on 2016/7/19

:author: hubo
'''



from vlcp.config.config import defaultconfig
from vlcp.server.module import depend, api, call_api
from vlcp.service.sdn.flowbase import FlowBase
from vlcp.service.sdn.ofpmanager import FlowInitialize
from vlcp.utils.flowupdater import FlowUpdater
from vlcp.utils.networkmodel import PhysicalPort, LogicalPort, LogicalNetwork,\
    LogicalNetworkMap, SubNet
import vlcp.service.kvdb.objectdb as objectdb
import vlcp.service.sdn.ofpportmanager as ofpportmanager
from vlcp.event.runnable import RoutineContainer
from vlcp.protocol.openflow.openflow import OpenflowConnectionStateEvent,\
    OpenflowAsyncMessageEvent, OpenflowErrorResultException
from vlcp.utils.ethernet import ethernet_l2, ip4_addr, mac_addr, mac_addr_bytes,\
    ip4_addr_bytes, ethernet_l7, ip4_packet_l7, ip4_payload, ip4_udp_payload
import vlcp.service.sdn.ioprocessing as iop
import itertools
from vlcp.utils.dataobject import ReferenceObject
from vlcp.service.sdn import arpresponder
import vlcp.utils.dhcp as d
from vlcp.utils.netutils import parse_ip4_network, get_broadcast, get_netmask
import os
from namedstruct.stdprim import uint16
from vlcp.event.event import M_

class DHCPUpdater(FlowUpdater):
    def __init__(self, connection, parent):
        FlowUpdater.__init__(self, connection, (), ('DHCPUpdater', connection), parent._logger)
        self._parent = parent
        self._lastlognets = ()
        self._lastlogports = ()
        self._lastlogportinfo = {}
        self._lastlognetinfo = {}
        self._lastserveraddresses = set()
        self._dhcpentries = {}

    async def main(self):
        try:
            if self._connection.protocol.disablenxext:
                self._logger.warning("DHCP disabled on connection %r because Nicira extension is not enabled", self._connection)
                return
            self.subroutine(self._update_handler(), True, '_update_handler_routine')
            self.subroutine(self._dhcp_handler(), True, '_dhcp_handler_routine')
            await FlowUpdater.main(self)
        finally:
            if hasattr(self, '_update_handler_routine'):
                self._update_handler_routine.close()
            if hasattr(self, '_dhcp_handler_routine'):
                self._dhcp_handler_routine.close()
    async def _dhcp_handler(self):
        """
        Mini DHCP server, respond DHCP packets from OpenFlow
        """
        conn = self._connection
        ofdef = self._connection.openflowdef
        l3 = self._parent._gettableindex('l3input', self._connection.protocol.vhost)
        dhcp_packet_matcher = OpenflowAsyncMessageEvent.createMatcher(ofdef.OFPT_PACKET_IN, None, None, l3, 1,
                                                              self._connection, self._connection.connmark)
        # These tags are important options. They are sent first to make sure the client
        # correctly receive these options.
        required_tags = [d.OPTION_MESSAGE_TYPE, d.OPTION_SERVER_IDENTIFIER,
                         d.OPTION_NETMASK, d.OPTION_ROUTER,
                         d.OPTION_DNSSERVER, d.OPTION_BROADCAST, d.OPTION_MTU,
                         d.OPTION_LEASE_TIME, d.OPTION_T1, d.OPTION_T2]
        server_mac = mac_addr(self._parent.servermac)
        # IP fragment identifier
        trans_id = uint16.create(os.urandom(2))
        def set_options(payload, option_dict, provide_options, message_type, remove_lease = False):
            """
            Set DHCP options to output payload regarding the incoming request
            
            :param payload: output DHCP payload
            
            :param option_dict: incoming DHCP options in request
            
            :param provide_options: all DHCP options that are ready to sent to the client
            
            :param message_type: output DHCP message type
            
            :param remove_lease: remove all leases options. DHCPINFORM cannot contain
                                 leases options. See https://tools.ietf.org/html/rfc2131#section-3.4
            """
            message_type_opt = d.dhcp_option_message_type(value = message_type)
            if d.OPTION_REQUESTED_OPTIONS in option_dict:
                # First requested options, then required options, then others
                reqs = set(option_dict[d.OPTION_REQUESTED_OPTIONS].value)
                send_tags = [t for t in option_dict[d.OPTION_REQUESTED_OPTIONS].value
                             if t == d.OPTION_MESSAGE_TYPE or t in provide_options] \
                            + [t for t in required_tags
                               if (t in provide_options or t == d.OPTION_MESSAGE_TYPE) and t not in reqs] \
                            + [t for t in provide_options
                               if t not in reqs and t not in required_tags]
            else:
                # Required options, then others
                send_tags = [t for t in required_tags
                             if t in provide_options or t == d.OPTION_MESSAGE_TYPE] \
                            + [t for t in set(provide_options.keys()).difference(required_tags)]
            # If the client has sent an option for max message size, use it; or use the RFC required 576 bytes
            not_finished = d.build_options(
                                payload,
                                [message_type_opt if t == d.OPTION_MESSAGE_TYPE
                                  else provide_options[t] for t in send_tags
                                  if not remove_lease or (t != d.OPTION_LEASE_TIME and t != d.OPTION_T1 and t != OPTION_T2)],
                                max(min(option_dict[d.OPTION_MAX_MESSAGE_SIZE].value, 1400), 576)
                                if d.OPTION_MAX_MESSAGE_SIZE in option_dict
                                else 576)
        async def send_packet(pid, packet):
            """
            Send DHCP packet to specified port
            """
            await self.execute_commands(conn,
                            [ofdef.ofp_packet_out(
                                    buffer_id = ofdef.OFP_NO_BUFFER,
                                    in_port = ofdef.OFPP_CONTROLLER,
                                    actions = [ofdef.ofp_action_output(port = pid,
                                                                       max_len = ofdef.OFPCML_NO_BUFFER)],
                                    data = packet._tobytes()
                                )])
        while True:
            ev = await dhcp_packet_matcher
            msg = ev.message
            try:
                in_port = ofdef.ofp_port_no.create(ofdef.get_oxm(msg.match.oxm_fields, ofdef.OXM_OF_IN_PORT))
                if in_port not in self._dhcpentries:
                    # Not a DHCP-enabled port
                    continue
                port_mac, port_ip, server_ip, provide_options = self._dhcpentries[in_port]
                # Fragmented DHCP packets are not supported - this is allowed according
                # to RFC: server and clients are only needed to support at least 576 bytes
                # DHCP messages.
                l7_packet = ethernet_l7.create(msg.data)
                dhcp_packet = d.dhcp_payload.create(l7_packet.data)
                # We only process a DHCP request directly sent from the logical port
                if (dhcp_packet.op != d.BOOTREQUEST                            # A DHCP server packet
                        or dhcp_packet.hlen != 6 or dhcp_packet.htype != 1     # Hardware address not ethernet (48-bit)
                        or dhcp_packet.magic_cookie != d.BOOTP_MAGIC_COOKIE    # Magic cookie not match
                        or dhcp_packet.giaddr != 0):                           # A relayed DHCP message
                    raise ValueError('Unsupported DHCP packet')
                # Reassemble DHCP options
                options = d.reassemble_options(dhcp_packet)
                option_dict = dict((o.tag, o) for o in options)
                if d.OPTION_MESSAGE_TYPE not in option_dict:
                    raise ValueError('Message type not found')
                message_type = option_dict[d.OPTION_MESSAGE_TYPE].value
                is_nak = False
                if message_type == d.DHCPDISCOVER:
                    # A DHCPDISCOVER should get a DHCPOFFER response
                    if dhcp_packet.chaddr[:6].ljust(6, b'\x00') != mac_addr.tobytes(port_mac):
                        # MAC address not match, ignore this packet
                        continue
                    dhcp_reply = d.dhcp_payload(op = d.BOOTREPLY,
                                                htype = 1,
                                                hlen = 6,
                                                hops = 0,
                                                xid = dhcp_packet.xid,
                                                secs = 0,
                                                flags = dhcp_packet.flags,
                                                ciaddr = 0,
                                                yiaddr = port_ip,
                                                siaddr = 0,
                                                giaddr = dhcp_packet.giaddr,
                                                chaddr = dhcp_packet.chaddr,
                                                magic_cookie = d.BOOTP_MAGIC_COOKIE
                                                )
                    set_options(dhcp_reply, option_dict, provide_options, d.DHCPOFFER)
                elif message_type == d.DHCPREQUEST:
                    # A DHCPREQUEST should get a DHCPACK reply
                    if d.OPTION_SERVER_IDENTIFIER in option_dict and option_dict[d.OPTION_SERVER_IDENTIFIER].value != server_ip:
                        # Ignore packets to wrong address
                        continue
                    if dhcp_packet.chaddr[:6].ljust(6, b'\x00') != mac_addr.tobytes(port_mac) \
                            or (d.OPTION_REQUESTED_IP in option_dict and option_dict[d.OPTION_REQUESTED_IP].value != port_ip) \
                            or (dhcp_packet.ciaddr != 0 and dhcp_packet.ciaddr != port_ip):
                        # Requested MAC or IP not matched, send a NACK
                        dhcp_reply = d.dhcp_payload(op = d.BOOTREPLY,
                                                    htype = 1,
                                                    hlen = 6,
                                                    hops = 0,
                                                    xid = dhcp_packet.xid,
                                                    secs = 0,
                                                    flags = dhcp_packet.flags,
                                                    ciaddr = 0,
                                                    yiaddr = 0,
                                                    siaddr = 0,
                                                    giaddr = dhcp_packet.giaddr,
                                                    chaddr = dhcp_packet.chaddr,
                                                    magic_cookie = d.BOOTP_MAGIC_COOKIE
                                                    )
                        # Do not send more options in NACK
                        d.build_options(dhcp_reply, [d.dhcp_option_message_type(value = d.DHCPNAK),
                                                     d.dhcp_option_address(tag = d.OPTION_SERVER_IDENTIFIER,
                                                                           value = server_ip)], 576, 0)
                        is_nak = True
                    else:
                        dhcp_reply = d.dhcp_payload(op = d.BOOTREPLY,
                                                    htype = 1,
                                                    hlen = 6,
                                                    hops = 0,
                                                    xid = dhcp_packet.xid,
                                                    secs = 0,
                                                    flags = dhcp_packet.flags,
                                                    ciaddr = dhcp_packet.ciaddr,
                                                    yiaddr = port_ip,
                                                    siaddr = 0,
                                                    giaddr = dhcp_packet.giaddr,
                                                    chaddr = dhcp_packet.chaddr,
                                                    magic_cookie = d.BOOTP_MAGIC_COOKIE
                                                    )
                        set_options(dhcp_reply, option_dict, provide_options, d.DHCPACK)
                elif message_type == d.DHCPDECLINE:
                    # Address already in use?
                    self._logger.warning('DHCP client reports DHCPDECLINE, there should be some problems.'\
                                         ' Connection = %r(%016x), port = %d.',
                                         self._connection, self._connection.openflow_datapathid)
                elif message_type == d.DHCPRELEASE:
                    # Safe to ignore, we do not use a dynamic IP address pool
                    continue
                elif message_type == d.DHCPINFORM:
                    # Client setup IP addresses itself, but requesting more information
                    # DHCPINFORM reply cannot have lease options, and yiaddr = 0
                    dhcp_reply = d.dhcp_payload(op = d.BOOTREPLY,
                                                htype = 1,
                                                hlen = 6,
                                                hops = 0,
                                                xid = dhcp_packet.xid,
                                                secs = 0,
                                                flags = dhcp_packet.flags,
                                                ciaddr = dhcp_packet.ciaddr,
                                                yiaddr = 0,
                                                siaddr = 0,
                                                giaddr = dhcp_packet.giaddr,
                                                chaddr = dhcp_packet.chaddr,
                                                magic_cookie = d.BOOTP_MAGIC_COOKIE
                                                )
                    set_options(dhcp_reply, option_dict, provide_options, d.DHCPACK, True)
                trans_id = (trans_id + 1) & 0xffff
                if (dhcp_packet.flags & d.DHCPFLAG_BROADCAST) or is_nak:
                    # client request broadcast, or DHCPNAK
                    # RFC requires that DHCPNAK uses broadcast
                    dl_dst = [0xff, 0xff, 0xff, 0xff, 0xff, 0xff]
                    ip_dst = 0xffffffff
                else:
                    dl_dst = l7_packet.dl_src
                    ip_dst = port_ip
                reply_packet = ip4_packet_l7((ip4_payload, ip4_udp_payload),
                                           dl_src = server_mac,
                                           dl_dst = dl_dst,
                                           identifier = trans_id,
                                           ttl = 128,
                                           ip_src = server_ip,
                                           ip_dst = ip_dst,
                                           sport = 67,
                                           dport = 68,
                                           data = dhcp_reply._tobytes()
                                           )
                # Send packet to the incoming port
                self.subroutine(send_packet(in_port, reply_packet), True)
            except Exception:
                self._logger.info('Invalid DHCP packet received: %r', msg.data, exc_info = True)
            
    async def _update_handler(self):
        dataobjectchanged = iop.DataObjectChanged.createMatcher(None, None, self._connection)
        while True:
            ev = await dataobjectchanged
            self._lastlogports, _, self._lastlognets, _ = ev.current
            self._update_walk()

    def _walk_logport(self, key, value, walk, save):
        if value is not None:
            save(key)
            if hasattr(value, 'subnet'):
                try:
                    _ = walk(value.subnet.getkey())
                except KeyError:
                    pass
                else:
                    save(value.subnet.getkey())

    def _walk_lognet(self, key, value, walk, save):
        save(key)
        if value is None:
            return

    def _update_walk(self):
        logport_keys = [p.getkey() for p,_ in self._lastlogports]
        lognet_keys = [n.getkey() for n,_ in self._lastlognets]
        lognet_mapkeys = [LogicalNetworkMap.default_key(n.id) for n,_ in self._lastlognets]
        self._initialkeys = logport_keys + lognet_keys + lognet_mapkeys
        self._walkerdict = dict(itertools.chain(((n, self._walk_lognet) for n in lognet_keys),
                                                ((p, self._walk_logport) for p in logport_keys)))
        self.subroutine(self.restart_walk(), False)

    async def updateflow(self, connection, addvalues, removevalues, updatedvalues):
        try:
            allobjs = set(o for o in self._savedresult if o is not None and not o.isdeleted())
            # Create DHCP entries
            lastlognetinfo = self._lastlognetinfo
            lastlogportinfo = self._lastlogportinfo
            lastserveraddresses = self._lastserveraddresses
            currentlognetinfo = dict((n, id) for n,id in self._lastlognets if n in allobjs)
            currentlogportinfo = dict((p, (id, currentlognetinfo[p.network], p.ip_address, p.mac_address, getattr(p.subnet, 'dhcp_server', self._parent.serveraddress)))
                                      for p, id in self._lastlogports
                                      if p in allobjs and p.network in currentlognetinfo and hasattr(p, 'subnet')
                                      and hasattr(p, 'mac_address') and hasattr(p, "ip_address"))
            currentserveraddresses = set((v[4], self._parent.servermac, p.network.id, True)
                                         for p,v in currentlogportinfo.items())
            self._lastlognetinfo = currentlognetinfo
            self._lastlogportinfo = currentlogportinfo
            self._lastserveraddresses = currentserveraddresses
            dhcp_entries = {}
            t1 = self._parent.t1
            t2 = self._parent.t2
            for p in self._savedresult:
                if p is not None and not p.isdeleted() and p.isinstance(LogicalPort) and hasattr(p, 'subnet') \
                        and hasattr(p, 'ip_address') and hasattr(p, 'mac_address'):
                    if p.network in currentlognetinfo and p in currentlogportinfo:
                        try:
                            pid, _, ip_address, mac_address, server_address = currentlogportinfo[p]
                            if getattr(p.subnet, 'dhcp_enabled', False):
                                continue
                            cidr = p.subnet.cidr
                            network, mask = parse_ip4_network(cidr)
                            # options from lowest priority to highest
                            # the latter options can override the earlier options
                            # options from default settings
                            entries = {d.OPTION_LEASE_TIME: (lambda x: d.DHCPTIME_INFINITE if x is None else x)(self._parent.leasetime)
                                       }
                            # options from network
                            if hasattr(p.network, 'dns_nameservers'):
                                entries[d.OPTION_DNSSERVER] = p.network.dns_nameservers
                            if hasattr(p.network, 'domain_name'):
                                entries[d.OPTION_DOMAINNAME] = p.network.domain_name
                            if hasattr(p.network, 'mtu'):
                                entries[d.OPTION_MTU] = p.network.mtu
                            if hasattr(p.network, 'ntp_servers'):
                                entries[d.OPTION_NTPSERVER] = p.network.ntp_servers
                            if hasattr(p.network, 'lease_time'):
                                entries[d.OPTION_LEASE_TIME] = p.network.lease_time
                            if hasattr(p.network, 'extra_dhcp_options'):
                                entries.update(p.network.extra_dhcp_options)
                            options = d.create_dhcp_options(entries, True, True)
                            # options from subnet
                            entries = {d.OPTION_NETMASK : ip4_addr.formatter(get_netmask(mask)),
                                       d.OPTION_BROADCAST : ip4_addr.formatter(get_broadcast(network, mask))}
                            if hasattr(p.subnet, 'gateway'):
                                entries[d.OPTION_ROUTER] = p.subnet.gateway
                            if hasattr(p.subnet, 'dns_nameservers'):
                                entries[d.OPTION_DNSSERVER] = p.subnet.dns_nameservers
                            if hasattr(p.subnet, 'domain_name'):
                                entries[d.OPTION_DOMAINNAME] = p.subnet.domain_name
                            if hasattr(p.subnet, 'mtu'):
                                entries[d.OPTION_MTU] = p.subnet.mtu
                            if hasattr(p.subnet, 'ntp_servers'):
                                entries[d.OPTION_NTPSERVER] = p.subnet.ntp_servers
                            if hasattr(p.subnet, 'lease_time'):
                                entries[d.OPTION_LEASE_TIME] = p.subnet.lease_time
                            # Routes is special
                            if hasattr(p.subnet, 'host_routes'):
                                routes = list(p.subnet.host_routes)
                                if routes and not any(parse_ip4_network(r[0]) == (0,0) for r in routes):
                                    # Add gateway as the default route
                                    routes.append(['0.0.0.0/0', p.subnet.gateway])
                                entries[d.OPTION_CLASSLESSROUTE] = routes
                            # TODO: add extra routes from routers
                            if hasattr(p.subnet, 'extra_dhcp_options'):
                                entries.update(p.subnet.extra_dhcp_options)
                            options.update(d.create_dhcp_options(entries, True, True))
                            entries = {}
                            # options from port
                            if hasattr(p, 'hostname'):
                                entries[d.OPTION_HOSTNAME] = p.hostname
                            if hasattr(p, 'extra_dhcp_options'):
                                entries.update(p.extra_dhcp_options)
                            options.update(d.create_dhcp_options(entries, True, True))
                            # server address
                            entries = {d.OPTION_SERVER_IDENTIFIER: server_address}
                            options.update(d.create_dhcp_options(entries, True, True))
                            # remove empty
                            options = dict((k,v) for k,v in options.items() if v is not None)
                            # set the lease time to infinite if not set
                            if d.OPTION_LEASE_TIME not in options:
                                options[d.OPTION_LEASE_TIME] = d.create_option_from_value(d.OPTION_LEASE_TIME, d.DHCPTIME_INFINITE)
                            # if lease time if not infinite, set T1 and T2 if not already set
                            if options[d.OPTION_LEASE_TIME].value != d.DHCPTIME_INFINITE:
                                leasetime = options[d.OPTION_LEASE_TIME].value
                                if d.OPTION_T1 not in options and t1 is not None and t1 < leasetime:
                                    options[d.OPTION_T1] = d.create_option_from_value(d.OPTION_T1, t1)
                                if d.OPTION_T2 not in options and t2 is not None and t2 < leasetime:
                                    options[d.OPTION_T1] = d.create_option_from_value(d.OPTION_T1, t2)
                            # Save DHCP entry
                            dhcp_entries[pid] = (mac_addr(mac_address), ip4_addr(ip_address),
                                                 ip4_addr(server_address), options)
                        except Exception:
                            self._logger.warning("Failed to create DHCP options for port id=%r. Will disable DHCP on this port",
                                                 p.id, exc_info = True)
            self._dhcpentries = dhcp_entries
            # Create flows
            # For every logical port, create two flows - one for unicast, one for broadcast
            ofdef = connection.openflowdef
            vhost = connection.protocol.vhost
            l3 = self._parent._gettableindex('l3input', vhost)
            cmds = []
            if connection.protocol.disablenxext:
                def match_network(nid):
                    return [ofdef.create_oxm(ofdef.OXM_OF_METADATA_W, ((nid & 0xffff) << 32) | (0xffff << 16),
                                            b'\x00\x00\xff\xff\x40\x00\x00\x00')]
            else:
                def match_network(nid):
                    return [ofdef.create_oxm(ofdef.NXM_NX_REG5, nid),
                            ofdef.create_oxm(ofdef.NXM_NX_REG7_W, 0x4000, 0x4000)]
            def _delete_flows(nid):
                return ofdef.ofp_flow_mod(cookie = 0x1,
                                           cookie_mask = 0xffffffffffffffff,
                                           table_id = l3,
                                           command = ofdef.OFPFC_DELETE,
                                           priority = ofdef.OFP_DEFAULT_PRIORITY,
                                           buffer_id = ofdef.OFP_NO_BUFFER,
                                           out_port = ofdef.OFPP_ANY,
                                           out_group = ofdef.OFPG_ANY,
                                           match = ofdef.ofp_match_oxm(
                                                        oxm_fields = 
                                                            match_network(nid) +
                                                            [ofdef.create_oxm(ofdef.OXM_OF_ETH_TYPE, ofdef.ETHERTYPE_IP),
                                                            ofdef.create_oxm(ofdef.OXM_OF_IP_PROTO, ofdef.IPPROTO_UDP),
                                                            ofdef.create_oxm(ofdef.OXM_OF_UDP_DST, 67)
                                                            ]
                                                    )
                                           )
            def _delete_flows2(nid, serveraddr):
                return ofdef.ofp_flow_mod(cookie = 0x1,
                                               cookie_mask = 0xffffffffffffffff,
                                               table_id = l3,
                                               command = ofdef.OFPFC_DELETE,
                                               priority = ofdef.OFP_DEFAULT_PRIORITY,
                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                               out_port = ofdef.OFPP_ANY,
                                               out_group = ofdef.OFPG_ANY,
                                               match = ofdef.ofp_match_oxm(
                                                            oxm_fields = 
                                                                match_network(nid) +
                                                                [ofdef.create_oxm(ofdef.OXM_OF_ETH_TYPE, ofdef.ETHERTYPE_IP),
                                                                ofdef.create_oxm(ofdef.OXM_OF_IP_PROTO, ofdef.IPPROTO_UDP),
                                                                ofdef.create_oxm(ofdef.OXM_OF_IPV4_DST,
                                                                                 ip4_addr_bytes(serveraddr)),
                                                                ofdef.create_oxm(ofdef.OXM_OF_UDP_DST, 67)
                                                                ]
                                                        )
                                               )
            def _create_flows(nid):
                return ofdef.ofp_flow_mod(cookie = 0x1,
                                           cookie_mask = 0xffffffffffffffff,
                                           table_id = l3,
                                           command = ofdef.OFPFC_ADD,
                                           priority = ofdef.OFP_DEFAULT_PRIORITY,
                                           buffer_id = ofdef.OFP_NO_BUFFER,
                                           out_port = ofdef.OFPP_ANY,
                                           out_group = ofdef.OFPG_ANY,
                                           match = ofdef.ofp_match_oxm(
                                                        oxm_fields = 
                                                            match_network(nid) +
                                                            [ofdef.create_oxm(ofdef.OXM_OF_ETH_DST_W, b'\x01\x00\x00\x00\x00\x00', b'\x01\x00\x00\x00\x00\x00'),
                                                            ofdef.create_oxm(ofdef.OXM_OF_ETH_TYPE, ofdef.ETHERTYPE_IP),
                                                            ofdef.create_oxm(ofdef.OXM_OF_IP_PROTO, ofdef.IPPROTO_UDP),
                                                            ofdef.create_oxm(ofdef.OXM_OF_UDP_DST, 67)
                                                            ]
                                                    ),
                                           instructions = [
                                                ofdef.ofp_instruction_actions(
                                                            actions = [
                                                                ofdef.ofp_action_output(port = ofdef.OFPP_CONTROLLER,
                                                                                        max_len = ofdef.OFPCML_NO_BUFFER
                                                                                        )
                                                            ]
                                                        )
                                                    ]
                                           )
            def _create_flows2(nid, serveraddr):
                return ofdef.ofp_flow_mod(cookie = 0x1,
                                           cookie_mask = 0xffffffffffffffff,
                                           table_id = l3,
                                           command = ofdef.OFPFC_ADD,
                                           priority = ofdef.OFP_DEFAULT_PRIORITY,
                                           buffer_id = ofdef.OFP_NO_BUFFER,
                                           out_port = ofdef.OFPP_ANY,
                                           out_group = ofdef.OFPG_ANY,
                                           match = ofdef.ofp_match_oxm(
                                                        oxm_fields = match_network(nid) +
                                                            [ofdef.create_oxm(ofdef.OXM_OF_ETH_TYPE, ofdef.ETHERTYPE_IP),
                                                            ofdef.create_oxm(ofdef.OXM_OF_IP_PROTO, ofdef.IPPROTO_UDP),
                                                            ofdef.create_oxm(ofdef.OXM_OF_IPV4_DST,
                                                                             ip4_addr_bytes(serveraddr)),
                                                            ofdef.create_oxm(ofdef.OXM_OF_UDP_DST, 67)
                                                            ]
                                                    ),
                                           instructions = [
                                                ofdef.ofp_instruction_actions(
                                                            actions = [
                                                                ofdef.ofp_action_output(port = ofdef.OFPP_CONTROLLER,
                                                                                        max_len = ofdef.OFPCML_NO_BUFFER
                                                                                        )
                                                            ]
                                                        )
                                                    ]
                                           )
            for n in removevalues:
                if n in lastlognetinfo:
                    nid = lastlognetinfo[n]
                    cmds.append(_delete_flows(nid))
            lastnetdict = dict((n.id,n) for n in lastlognetinfo)
            for serveraddr in lastserveraddresses:
                if serveraddr not in currentserveraddresses:
                    addr, _, networkid, _ = serveraddr
                    if networkid in lastnetdict:
                        n = lastnetdict[networkid]
                        nid = lastlognetinfo[n]
                        cmds.append(_delete_flows2(nid, addr))
            await self.execute_commands(connection, cmds)
            
            # Remove old ARP entries; Add new ARP entries
            remove_arps = lastserveraddresses.difference(currentserveraddresses)
            if remove_arps:
                await call_api(self, 'arpresponder', 'removeproxyarp', {'connection': connection,
                                                                          'arpentries': remove_arps})
            add_arps = currentserveraddresses.difference(lastserveraddresses)
            if add_arps:
                await call_api(self, 'arpresponder', 'createproxyarp', {'connection': connection,
                                                                          'arpentries': add_arps})
            del cmds[:]
            for n in addvalues:
                if n in currentlognetinfo:
                    nid = currentlognetinfo[n]
                    cmds.append(_create_flows(nid))
            currnetdict = dict((n.id,n) for n in currentlognetinfo)
            for serveraddr in currentserveraddresses:
                if serveraddr not in lastserveraddresses:
                    addr, _, networkid, _ = serveraddr
                    if networkid in currnetdict:
                        n = currnetdict[networkid]
                        nid = currentlognetinfo[n]
                        cmds.append(_create_flows2(nid, addr))
            await self.execute_commands(connection, cmds)
        except Exception:
            self._logger.warning("Unexpected exception in DHCPUpdater. Will ignore and continue.", exc_info = True)



@defaultconfig
@depend(ofpportmanager.OpenflowPortManager, objectdb.ObjectDB, arpresponder.ARPResponder)
class DHCPServer(FlowBase):
    "DHCP server that responds the DHCP discover/request with static IP address settings"
    _tablerequest = (("l3input", ('l2input',), ''),
                     ("l2output", ("l3input",), ''))
    # Responding DHCP server address
    _default_serveraddress = '169.254.169.254'
    # Responding DHCP server MAC address
    _default_servermac = '1a:23:67:59:63:33'
    # DHCP leases timeout time
    _default_leasetime = None
    # DHCP default T1 option
    _default_t1 = None
    # DHCP default T2 option
    _default_t2 = None
    def __init__(self, server):
        FlowBase.__init__(self, server)
        self.apiroutine = RoutineContainer(self.scheduler)
        self.apiroutine.main = self._main
        self.routines.append(self.apiroutine)
        self._flowupdaters = {}
        self._extra_arps = {}

    async def _main(self):
        flow_init = FlowInitialize.createMatcher(_ismatch = lambda x: self.vhostbind is None or x.vhost in self.vhostbind)
        conn_down = OpenflowConnectionStateEvent.createMatcher(state = OpenflowConnectionStateEvent.CONNECTION_DOWN,
                                                               _ismatch = lambda x: self.vhostbind is None or x.createby.vhost in self.vhostbind)
        while True:
            ev, m = await M_(flow_init, conn_down)
            if m is flow_init:
                c = ev.connection
                self.apiroutine.subroutine(self._init_conn(c))
            else:
                c = ev.connection
                self.apiroutine.subroutine(self._remove_conn(c))

    async def _init_conn(self, conn):
        # Default
        if conn in self._flowupdaters:
            updater = self._flowupdaters.pop(conn)
            updater.close()
        updater = DHCPUpdater(conn, self)
        #flowupdater = VXLANFlowUpdater(conn, self)
        self._flowupdaters[conn] = updater
        updater.start()

    async def _remove_conn(self, conn):
        # Do not need to modify flows
        if conn in self._flowupdaters:
            self._flowupdaters.pop(conn).close()



"""
Physical network with type=native does not have isolation technique,
so there can be only one logical network in each physical network.
"""
from vlcp.server.module import Module, publicapi
from vlcp.event.runnable import RoutineContainer
from vlcp.config.config import defaultconfig
from vlcp.utils.networkplugin import createphysicalnetwork,\
    updatephysicalnetwork, default_physicalnetwork_keys, deletephysicalnetwork,\
    deletephysicalport, default_physicalport_keys, createlogicalnetwork,\
    default_logicalnetwork_keys, default_processor, updatelogicalnetwork,\
    deletelogicalnetwork, createphysicalport, updatephysicalport

@defaultconfig
class NetworkNativeDriver(Module):
    """
    Network driver for native networks. Native network is a physical network
    provides only one logical network capacity. Packets from the logical network
    is directly forwarded to the physical network.
    """
    def __init__(self,server):
        super(NetworkNativeDriver,self).__init__(server)
        self.app_routine = RoutineContainer(self.scheduler)
        self.app_routine.main = self._main
        self.routines.append(self.app_routine)
        self.createAPI(
                       publicapi(self.createphysicalnetwork,
                                    criteria=lambda type: type == 'native'),
                       publicapi(self.updatephysicalnetwork,
                                    criteria=lambda type: type == 'native'),
                       publicapi(self.deletephysicalnetwork,
                                    criteria=lambda type: type == 'native'),
                       publicapi(self.createphysicalport,
                                    criteria=lambda type: type == 'native'),
                       publicapi(self.updatephysicalport,
                                    criteria=lambda type: type == 'native'),
                       publicapi(self.deletephysicalport,
                                    criteria=lambda type: type == 'native'),
                       publicapi(self.createlogicalnetwork,
                                    criteria=lambda type: type == 'native'),
                       publicapi(self.updatelogicalnetwork,
                                    criteria=lambda type: type == 'native'),
                       publicapi(self.deletelogicalnetwork,
                                    criteria=lambda type: type == "native"),
                       #used in IOprocessing module
                       publicapi(self.createioflowparts,
                                    criteria=lambda connection,logicalnetwork,
                                    physicalport,logicalnetworkid,physicalportid:
                                    logicalnetwork.physicalnetwork.type == "native")

                       )

    async def _main(self):
        self._logger.info("network_native_driver running ---")

    def createphysicalnetwork(self, type):
        # create an new physical network
        return createphysicalnetwork(type), default_physicalnetwork_keys
    
    def updatephysicalnetwork(self, type):
        # update a physical network
        return updatephysicalnetwork(), None
    
    def deletephysicalnetwork(self, type):
        return deletephysicalnetwork(), default_physicalnetwork_keys
    
    def createphysicalport(self, type):
        return createphysicalport(), default_physicalport_keys
        
    def updatephysicalport(self, type):
        return updatephysicalport(), None
   
    def deletephysicalport(self, type):
        return deletephysicalport(), default_physicalport_keys

    def createlogicalnetwork(self, type):
        def logicalnetwork_processor(logicalnetwork, logicalnetworkmap, physicalnetwork,
                                     physicalnetworkmap, walk, write, *, parameters):
            if physicalnetworkmap.logicnetworks.dataset():
                raise ValueError("physical network with type=native can only have one logical network")
            return default_processor(logicalnetwork, parameters=parameters, excluding=('id', 'physicalnetwork'))
        return createlogicalnetwork(create_processor=logicalnetwork_processor), default_logicalnetwork_keys

    def updatelogicalnetwork(self, type):
        return updatelogicalnetwork(), None

    def deletelogicalnetwork(self, type):
        return deletelogicalnetwork(), default_logicalnetwork_keys

    def createioflowparts(self,connection,logicalnetwork,physicalport,logicalnetworkid,physicalportid):

        #
        #  1. used in IOProcessing , when physicalport add to logicalnetwork 
        #     return : input flow match vxlan vni, input flow vlan parts actions
        #              output flow vxlan parts actions, output group bucket
        #
        
        input_match_oxm = []
        input_action = []
        output_action = [
                    connection.openflowdef.ofp_action_output(
                           port = physicalportid 
                        )
                ]

        output_group_bucket_action = [
                    connection.openflowdef.ofp_action_output(
                            port = physicalportid
                        )
                ]
        output_action2 = [
                    connection.openflowdef.ofp_action_output(
                           port = connection.openflowdef.OFPP_IN_PORT
                        )
                ]
        return input_match_oxm,input_action,output_action,output_group_bucket_action,output_action2



'''
Created on 2016/2/19

:author: hubo
'''

from vlcp.config import defaultconfig
from vlcp.server.module import Module, api, depend, call_api, ModuleNotification
from vlcp.event.runnable import RoutineContainer
from vlcp.service.connection import openflowserver
from vlcp.protocol.openflow import OpenflowConnectionStateEvent
from vlcp.event.connection import ConnectionResetException, ResolveRequestEvent,\
    ResolveResponseEvent
import itertools
import socket
from vlcp.event.event import Event, withIndices, M_
from vlcp.event.core import QuitException, syscall_removequeue
from contextlib import closing

def _get_endpoint(conn):
    raddr = getattr(conn, 'remoteaddr', None)
    if raddr:
        if isinstance(raddr, tuple):
            # Ignore port
            return raddr[0]
        else:
            # Unix socket
            return raddr
    else:
        return ''

@withIndices()
class TableAcquireUpdate(Event):
    pass

@withIndices('connection', 'datapathid', 'vhost')
class FlowInitialize(Event):
    pass

@withIndices()
class TableAcquireDelayEvent(Event):
    pass

@defaultconfig
@depend(openflowserver.OpenflowServer)
class OpenflowManager(Module):
    '''
    Manage Openflow Connections
    '''
    service = True
    _default_vhostbind = None
    def __init__(self, server):
        Module.__init__(self, server)
        self.apiroutine = RoutineContainer(self.scheduler)
        self.apiroutine.main = self._manage_conns
        self.routines.append(self.apiroutine)
        self.managed_conns = {}
        self.endpoint_conns = {}
        self.table_modules = set()
        self._acquiring = False
        self._acquire_updated = False
        self._lastacquire = None
        self._synchronized = False
        self.createAPI(api(self.getconnections, self.apiroutine),
                       api(self.getconnection, self.apiroutine),
                       api(self.waitconnection, self.apiroutine),
                       api(self.getdatapathids, self.apiroutine),
                       api(self.getalldatapathids, self.apiroutine),
                       api(self.getallconnections, self.apiroutine),
                       api(self.getconnectionsbyendpoint, self.apiroutine),
                       api(self.getconnectionsbyendpointname, self.apiroutine),
                       api(self.getendpoints, self.apiroutine),
                       api(self.getallendpoints, self.apiroutine),
                       api(self.acquiretable, self.apiroutine),
                       api(self.unacquiretable, self.apiroutine),
                       api(self.lastacquiredtables)
                       )

    def _add_connection(self, conn):
        vhost = conn.protocol.vhost
        conns = self.managed_conns.setdefault((vhost, conn.openflow_datapathid), [])
        remove = []
        for i in range(0, len(conns)):
            if conns[i].openflow_auxiliaryid == conn.openflow_auxiliaryid:
                ci = conns[i]
                remove = [ci]
                ep = _get_endpoint(ci)
                econns = self.endpoint_conns.get((vhost, ep))
                if econns is not None:
                    try:
                        econns.remove(ci)
                    except ValueError:
                        pass
                    if not econns:
                        del self.endpoint_conns[(vhost, ep)]
                del conns[i]
                break
        conns.append(conn)
        ep = _get_endpoint(conn)
        econns = self.endpoint_conns.setdefault((vhost, ep), [])
        econns.append(conn)
        if self._lastacquire and conn.openflow_auxiliaryid == 0:
            self.apiroutine.subroutine(self._initialize_connection(conn))
        return remove

    async def _initialize_connection(self, conn):
        ofdef = conn.openflowdef
        flow_mod = ofdef.ofp_flow_mod(buffer_id = ofdef.OFP_NO_BUFFER,
                                                 out_port = ofdef.OFPP_ANY,
                                                 command = ofdef.OFPFC_DELETE
                                                 )
        if hasattr(ofdef, 'OFPG_ANY'):
            flow_mod.out_group = ofdef.OFPG_ANY
        if hasattr(ofdef, 'OFPTT_ALL'):
            flow_mod.table_id = ofdef.OFPTT_ALL
        if hasattr(ofdef, 'ofp_match_oxm'):
            flow_mod.match = ofdef.ofp_match_oxm()
        cmds = [flow_mod]
        if hasattr(ofdef, 'ofp_group_mod'):
            group_mod = ofdef.ofp_group_mod(command = ofdef.OFPGC_DELETE,
                                            group_id = ofdef.OFPG_ALL
                                            )
            cmds.append(group_mod)
        await conn.protocol.batch(cmds, conn, self.apiroutine)
        if hasattr(ofdef, 'ofp_instruction_goto_table'):
            # Create default flows
            vhost = conn.protocol.vhost
            if self._lastacquire and vhost in self._lastacquire:
                _, pathtable = self._lastacquire[vhost]
                cmds = [ofdef.ofp_flow_mod(table_id = t[i][1],
                                             command = ofdef.OFPFC_ADD,
                                             priority = 0,
                                             buffer_id = ofdef.OFP_NO_BUFFER,
                                             out_port = ofdef.OFPP_ANY,
                                             out_group = ofdef.OFPG_ANY,
                                             match = ofdef.ofp_match_oxm(),
                                             instructions = [ofdef.ofp_instruction_goto_table(table_id = t[i+1][1])]
                                       )
                          for _,t in pathtable.items()
                          for i in range(0, len(t) - 1)]
                if cmds:
                    await conn.protocol.batch(cmds, conn, self.apiroutine)
        await self.apiroutine.wait_for_send(FlowInitialize(conn, conn.openflow_datapathid, conn.protocol.vhost))

    async def _acquire_tables(self):
        try:
            while self._acquire_updated:
                result = None
                exception = None
                # Delay the update so we are not updating table acquires for every module
                await self.apiroutine.wait_for_send(TableAcquireDelayEvent())
                await TableAcquireDelayEvent.createMatcher()
                module_list = list(self.table_modules)
                self._acquire_updated = False
                try:
                    requests = await self.apiroutine.execute_all(call_api(self.apiroutine, module, 'gettablerequest', {})
                                                                 for module in module_list)
                except QuitException:
                    raise
                except Exception as exc:
                    self._logger.exception('Acquiring table failed')
                    exception = exc
                else:
                    vhosts = set(vh for _, vhs in requests if vhs is not None for vh in vhs)
                    vhost_result = {}
                    # Requests should be list of (name, (ancester, ancester, ...), pathname)
                    for vh in vhosts:
                        graph = {}
                        table_path = {}
                        try:
                            for r in requests:
                                if r[1] is None or vh in r[1]:
                                    for name, ancesters, pathname in r[0]:
                                        if name in table_path:
                                            if table_path[name] != pathname:
                                                raise ValueError("table conflict detected: %r can not be in two path: %r and %r" % (name, table_path[name], pathname))
                                        else:
                                            table_path[name] = pathname
                                        if name not in graph:
                                            graph[name] = (set(ancesters), set())
                                        else:
                                            graph[name][0].update(ancesters)
                                        for anc in ancesters:
                                            graph.setdefault(anc, (set(), set()))[1].add(name)
                        except ValueError as exc:
                            self._logger.error(str(exc))
                            exception = exc
                            break
                        else:
                            sequences = []
                            def dfs_sort(current):
                                sequences.append(current)
                                for d in graph[current][1]:
                                    anc = graph[d][0]
                                    anc.remove(current)
                                    if not anc:
                                        dfs_sort(d)
                            nopre_tables = sorted([k for k,v in graph.items() if not v[0]], key = lambda x: (table_path.get(name, ''),name))
                            for t in nopre_tables:
                                dfs_sort(t)
                            if len(sequences) < len(graph):
                                rest_tables = set(graph.keys()).difference(sequences)
                                self._logger.error("Circle detected in table acquiring, following tables are related: %r, vhost = %r", sorted(rest_tables), vh)
                                self._logger.error("Circle dependencies are: %s", ", ".join(repr(tuple(graph[t][0])) + "=>" + t for t in rest_tables))
                                exception = ValueError("Circle detected in table acquiring, following tables are related: %r, vhost = %r" % (sorted(rest_tables),vh))
                                break
                            elif len(sequences) > 255:
                                self._logger.error("Table limit exceeded: %d tables (only 255 allowed), vhost = %r", len(sequences), vh)
                                exception = ValueError("Table limit exceeded: %d tables (only 255 allowed), vhost = %r" % (len(sequences),vh))
                                break
                            else:
                                full_indices = list(zip(sequences, itertools.count()))
                                tables = dict((k,tuple(g)) for k,g in itertools.groupby(sorted(full_indices, key = lambda x: table_path.get(x[0], '')),
                                                           lambda x: table_path.get(x[0], '')))
                                vhost_result[vh] = (full_indices, tables)
        finally:
            self._acquiring = False
        if exception:
            await self.apiroutine.wait_for_send(TableAcquireUpdate(exception = exception))
        else:
            result = vhost_result
            if result != self._lastacquire:
                self._lastacquire = result
                self._reinitall()
            await self.apiroutine.wait_for_send(TableAcquireUpdate(result = result))

    async def load(self, container):
        self.scheduler.queue.addSubQueue(1, TableAcquireDelayEvent.createMatcher(), 'ofpmanager_tableacquiredelay')
        await container.wait_for_send(TableAcquireUpdate(result = None))
        return await Module.load(self, container)

    async def unload(self, container, force=False):
        await Module.unload(self, container, force=force)
        await container.syscall(syscall_removequeue(self.scheduler.queue, 'ofpmanager_tableacquiredelay'))

    def _reinitall(self):
        for cl in self.managed_conns.values():
            for c in cl:
                self.apiroutine.subroutine(self._initialize_connection(c))

    async def _manage_existing(self):
        result = await call_api(self.apiroutine, "openflowserver", "getconnections", {})
        vb = self.vhostbind
        for c in result:
            if vb is None or c.protocol.vhost in vb:
                self._add_connection(c)
        self._synchronized = True
        await self.apiroutine.wait_for_send(ModuleNotification(self.getServiceName(), 'synchronized'))

    async def _wait_for_sync(self):
        if not self._synchronized:
            await ModuleNotification.createMatcher(self.getServiceName(), 'synchronized')

    async def _manage_conns(self):
        vb = self.vhostbind
        self.apiroutine.subroutine(self._manage_existing(), False)
        try:
            if vb is not None:
                conn_up = OpenflowConnectionStateEvent.createMatcher(state = OpenflowConnectionStateEvent.CONNECTION_SETUP,
                                                                     _ismatch = lambda x: x.createby.vhost in vb)
                conn_down = OpenflowConnectionStateEvent.createMatcher(state = OpenflowConnectionStateEvent.CONNECTION_DOWN,
                                                                     _ismatch = lambda x: x.createby.vhost in vb)
            else:
                conn_up = OpenflowConnectionStateEvent.createMatcher(state = OpenflowConnectionStateEvent.CONNECTION_SETUP)
                conn_down = OpenflowConnectionStateEvent.createMatcher(state = OpenflowConnectionStateEvent.CONNECTION_DOWN)
            while True:
                ev, m = await M_(conn_up, conn_down)
                if m is conn_up:
                    remove = self._add_connection(ev.connection)
                    self.scheduler.emergesend(ModuleNotification(self.getServiceName(), 'update', add = [ev.connection], remove = remove))
                else:
                    conns = self.managed_conns.get((ev.createby.vhost, ev.datapathid))
                    remove = []
                    if conns is not None:
                        try:
                            conns.remove(ev.connection)
                        except ValueError:
                            pass
                        else:
                            remove.append(ev.connection)
                        
                        if not conns:
                            del self.managed_conns[(ev.createby.vhost, ev.datapathid)]
                        # Also delete from endpoint_conns
                        ep = _get_endpoint(ev.connection)
                        econns = self.endpoint_conns.get((ev.createby.vhost, ep))
                        if econns is not None:
                            try:
                                econns.remove(ev.connection)
                            except ValueError:
                                pass
                            if not econns:
                                del self.endpoint_conns[(ev.createby.vhost, ep)]
                    if remove:
                        self.scheduler.emergesend(ModuleNotification(self.getServiceName(), 'update', add = [], remove = remove))
        finally:
            self.scheduler.emergesend(ModuleNotification(self.getServiceName(), 'unsynchronized'))

    async def getconnections(self, datapathid, vhost = ''):
        "Return all connections of datapath"
        await self._wait_for_sync()
        return list(self.managed_conns.get((vhost, datapathid), []))

    async def getconnection(self, datapathid, auxiliaryid = 0, vhost = ''):
        "Get current connection of datapath"
        await self._wait_for_sync()
        return self._getconnection(datapathid, auxiliaryid, vhost)
    def _getconnection(self, datapathid, auxiliaryid = 0, vhost = ''):
        conns = self.managed_conns.get((vhost, datapathid))
        if conns is None:
            return None
        else:
            for c in conns:
                if c.openflow_auxiliaryid == auxiliaryid:
                    return c
            return None

    async def waitconnection(self, datapathid, auxiliaryid = 0, timeout = 30, vhost = ''):
        "Wait for a datapath connection"
        await self._wait_for_sync()
        c = self._getconnection(datapathid, auxiliaryid, vhost)
        if c is None:
            timeout_, ev, _ = await self.apiroutine.wait_with_timeout(
                                                    timeout, 
                                                    OpenflowConnectionStateEvent.createMatcher(datapathid, auxiliaryid,
                                                            OpenflowConnectionStateEvent.CONNECTION_SETUP,
                                                            _ismatch = lambda x: x.createby.vhost == vhost))
            if timeout_:
                raise ConnectionResetException('Datapath %016x is not connected' % datapathid)
            return ev.connection
        else:
            return c

    async def getdatapathids(self, vhost = ''):
        "Get All datapath IDs"
        await self._wait_for_sync()
        return [k[1] for k in self.managed_conns.keys() if k[0] == vhost]

    async def getalldatapathids(self):
        "Get all datapath IDs from any vhost. Return ``(vhost, datapathid)`` pair."
        await self._wait_for_sync()
        return list(self.managed_conns.keys())

    async def getallconnections(self, vhost = ''):
        "Get all connections from vhost. If vhost is None, return all connections from any host"
        await self._wait_for_sync()
        if vhost is None:
            return list(itertools.chain(self.managed_conns.values()))
        else:
            return list(itertools.chain(v for k,v in self.managed_conns.items() if k[0] == vhost))

    async def getconnectionsbyendpoint(self, endpoint, vhost = ''):
        "Get connection by endpoint address (IP, IPv6 or UNIX socket address)"
        await self._wait_for_sync()
        return self.endpoint_conns.get((vhost, endpoint))

    async def getconnectionsbyendpointname(self, name, vhost = '', timeout = 30):
        "Get connection by endpoint name (Domain name, IP or IPv6 address)"
        # Resolve the name
        if not name:
            endpoint = ''
            return await self.getconnectionbyendpoint(endpoint, vhost)
        else:
            request = (name, 0, socket.AF_UNSPEC, socket.SOCK_STREAM, socket.IPPROTO_TCP, socket.AI_ADDRCONFIG | socket.AI_V4MAPPED)
            # Resolve hostname
            await self.apiroutine.wait_for_send(ResolveRequestEvent(request))
            timeout_, ev, m = await self.apiroutine.wait_with_timeout(timeout, ResolveResponseEvent.createMatcher(request))
            if timeout_:
                # Resolve is only allowed through asynchronous resolver
                #try:
                #    self.addrinfo = socket.getaddrinfo(self.hostname, self.port, socket.AF_UNSPEC, socket.SOCK_DGRAM if self.udp else socket.SOCK_STREAM, socket.IPPROTO_UDP if self.udp else socket.IPPROTO_TCP, socket.AI_ADDRCONFIG|socket.AI_NUMERICHOST)
                #except:
                raise IOError('Resolve hostname timeout: ' + name)
            else:
                if hasattr(ev, 'error'):
                    raise IOError('Cannot resolve hostname: ' + name)
                resp = ev.response
                for r in resp:
                    raddr = r[4]
                    if isinstance(raddr, tuple):
                        # Ignore port
                        endpoint = raddr[0]
                    else:
                        # Unix socket? This should not happen, but in case...
                        endpoint = raddr
                    r = await self.getconnectionsbyendpoint(endpoint, vhost)
                    if r is not None:
                        return r
            return None

    async def getendpoints(self, vhost = ''):
        "Get all endpoints for vhost"
        await self._wait_for_sync()
        return [k[1] for k in self.endpoint_conns if k[0] == vhost]

    async def getallendpoints(self):
        "Get all endpoints from any vhost. Return ``(vhost, endpoint)`` pairs."
        await self._wait_for_sync()
        return list(self.endpoint_conns.keys())

    def lastacquiredtables(self, vhost = ""):
        "Get acquired table IDs"
        return self._lastacquire.get(vhost)

    async def acquiretable(self, modulename):
        "Start to acquire tables for a module on module loading."
        if not modulename in self.table_modules:
            self.table_modules.add(modulename)
            self._acquire_updated = True
            if not self._acquiring:
                self._acquiring = True
                self.apiroutine.subroutine(self._acquire_tables())

    async def unacquiretable(self, modulename):
        "When module is unloaded, stop acquiring tables for this module."
        if modulename in self.table_modules:
            self.table_modules.remove(modulename)
            self._acquire_updated = True
            if not self._acquiring:
                self._acquiring = True
                self.apiroutine.subroutine(self._acquire_tables())



'''
Created on 2016/2/19

:author: hubo
'''

from vlcp.config import defaultconfig
from vlcp.server.module import Module, api, depend, call_api, ModuleNotification
from vlcp.event.runnable import RoutineContainer
from vlcp.service.connection import jsonrpcserver
from vlcp.protocol.jsonrpc import JsonRPCConnectionStateEvent,\
    JsonRPCProtocolException, JsonRPCErrorResultException,\
    JsonRPCNotificationEvent
from vlcp.event.connection import ConnectionResetException, ResolveRequestEvent,\
    ResolveResponseEvent
from vlcp.event.event import Event, withIndices, M_
from vlcp.utils import ovsdb
import socket
from contextlib import closing

@withIndices('systemid', 'connection', 'connmark', 'vhost')
class OVSDBConnectionSetup(Event):
    pass

@withIndices('state', 'datapathid', 'systemid', 'name', 'connection', 'connmark', 'vhost', 'bridgeuuid')
class OVSDBBridgeSetup(Event):
    UP = 'up'
    DOWN = 'down'

class OVSDBBridgeNotAppearException(Exception):
    pass

def _get_endpoint(conn):
    raddr = getattr(conn, 'remoteaddr', None)
    if raddr:
        if isinstance(raddr, tuple):
            # Ignore port
            return raddr[0]
        else:
            # Unix socket
            return raddr
    else:
        return ''


@defaultconfig
@depend(jsonrpcserver.JsonRPCServer)
class OVSDBManager(Module):
    '''
    Manage Openflow Connections
    '''
    service = True
    # Bind to JsonRPCServer vHosts. If not None, should be a list of vHost names e.g. ``['']``
    _default_vhostbind = None
    # Only acquire information from bridges with this names
    _default_bridgenames = None
    def __init__(self, server):
        Module.__init__(self, server)
        self.apiroutine = RoutineContainer(self.scheduler)
        self.apiroutine.main = self._manage_conns
        self.routines.append(self.apiroutine)
        self.managed_conns = {}
        self.managed_systemids = {}
        self.managed_bridges = {}
        self.managed_routines = []
        self.endpoint_conns = {}
        self.createAPI(api(self.getconnection, self.apiroutine),
                       api(self.waitconnection, self.apiroutine),
                       api(self.getdatapathids, self.apiroutine),
                       api(self.getalldatapathids, self.apiroutine),
                       api(self.getallconnections, self.apiroutine),
                       api(self.getbridges, self.apiroutine),
                       api(self.getbridge, self.apiroutine),
                       api(self.getbridgebyuuid, self.apiroutine),
                       api(self.waitbridge, self.apiroutine),
                       api(self.waitbridgebyuuid, self.apiroutine),
                       api(self.getsystemids, self.apiroutine),
                       api(self.getallsystemids, self.apiroutine),
                       api(self.getconnectionbysystemid, self.apiroutine),
                       api(self.waitconnectionbysystemid, self.apiroutine),
                       api(self.getconnectionsbyendpoint, self.apiroutine),
                       api(self.getconnectionsbyendpointname, self.apiroutine),
                       api(self.getendpoints, self.apiroutine),
                       api(self.getallendpoints, self.apiroutine),
                       api(self.getallbridges, self.apiroutine),
                       api(self.getbridgeinfo, self.apiroutine),
                       api(self.waitbridgeinfo, self.apiroutine)
                       )
        self._synchronized = False
    async def _update_bridge(self, connection, protocol, bridge_uuid, vhost):
        try:
            method, params = ovsdb.transact('Open_vSwitch',
                                            ovsdb.wait('Bridge', [["_uuid", "==", ovsdb.uuid(bridge_uuid)]],
                                                        ["datapath_id"], [{"datapath_id": ovsdb.oset()}], False, 5000),
                                            ovsdb.select('Bridge', [["_uuid", "==", ovsdb.uuid(bridge_uuid)]],
                                                                         ["datapath_id","name"]))
            jsonrpc_result, _ = await protocol.querywithreply(method, params, connection, self.apiroutine)
            r = jsonrpc_result[0]
            if 'error' in r:
                raise JsonRPCErrorResultException('Error while acquiring datapath-id: ' + repr(r['error']))
            r = jsonrpc_result[1]
            if 'error' in r:
                raise JsonRPCErrorResultException('Error while acquiring datapath-id: ' + repr(r['error']))
            if r['rows']:
                r0 = r['rows'][0]
                name = r0['name']
                dpid = int(r0['datapath_id'], 16)
                if self.bridgenames is None or name in self.bridgenames:
                    self.managed_bridges[connection].append((vhost, dpid, name, bridge_uuid))
                    self.managed_conns[(vhost, dpid)] = connection
                    await self.apiroutine.wait_for_send(OVSDBBridgeSetup(OVSDBBridgeSetup.UP,
                                                               dpid,
                                                               connection.ovsdb_systemid,
                                                               name,
                                                               connection,
                                                               connection.connmark,
                                                               vhost,
                                                               bridge_uuid))
        except JsonRPCProtocolException:
            pass

    async def _get_bridges(self, connection, protocol):
        try:
            try:
                vhost = protocol.vhost
                if not hasattr(connection, 'ovsdb_systemid'):
                    method, params = ovsdb.transact('Open_vSwitch', ovsdb.select('Open_vSwitch', [], ['external_ids']))
                    jsonrpc_result, _ = await protocol.querywithreply(method, params, connection, self.apiroutine)
                    result = jsonrpc_result[0]
                    system_id = ovsdb.omap_getvalue(result['rows'][0]['external_ids'], 'system-id')
                    connection.ovsdb_systemid = system_id
                else:
                    system_id = connection.ovsdb_systemid
                if (vhost, system_id) in self.managed_systemids:
                    oc = self.managed_systemids[(vhost, system_id)]
                    ep = _get_endpoint(oc)
                    econns = self.endpoint_conns.get((vhost, ep))
                    if econns:
                        try:
                            econns.remove(oc)
                        except ValueError:
                            pass
                    del self.managed_systemids[(vhost, system_id)]
                self.managed_systemids[(vhost, system_id)] = connection
                self.managed_bridges[connection] = []
                ep = _get_endpoint(connection)
                self.endpoint_conns.setdefault((vhost, ep), []).append(connection)
                method, params = ovsdb.monitor('Open_vSwitch', 'ovsdb_manager_bridges_monitor', {'Bridge':ovsdb.monitor_request(['name', 'datapath_id'])})
                try:
                    jsonrpc_result, _ = await protocol.querywithreply(method, params, connection, self.apiroutine)
                except JsonRPCErrorResultException:
                    # The monitor is already set, cancel it first
                    method, params = ovsdb.monitor_cancel('ovsdb_manager_bridges_monitor')
                    await protocol.querywithreply(method, params, connection, self.apiroutine, False)
                    method, params = ovsdb.monitor('Open_vSwitch', 'ovsdb_manager_bridges_monitor', {'Bridge':ovsdb.monitor_request(['name', 'datapath_id'])})
                    jsonrpc_result, _ = await protocol.querywithreply(method, params, connection, self.apiroutine)
            except Exception:
                await self.apiroutine.wait_for_send(OVSDBConnectionSetup(system_id, connection, connection.connmark, vhost))
                raise
            else:
                # Process initial bridges
                init_subprocesses = []
                if jsonrpc_result and 'Bridge' in jsonrpc_result:
                    init_subprocesses = [self._update_bridge(connection, protocol, buuid, vhost)
                                        for buuid in jsonrpc_result['Bridge'].keys()]
                async def init_process():
                    try:
                        await self.apiroutine.execute_all(init_subprocesses)
                    except Exception:
                        await self.apiroutine.wait_for_send(OVSDBConnectionSetup(system_id, connection, connection.connmark, vhost))
                        raise
                    else:
                        await self.apiroutine.waitForSend(OVSDBConnectionSetup(system_id, connection, connection.connmark, vhost))
                self.apiroutine.subroutine(init_process())
            # Wait for notify
            notification = JsonRPCNotificationEvent.createMatcher('update', connection, connection.connmark, _ismatch = lambda x: x.params[0] == 'ovsdb_manager_bridges_monitor')
            conn_down = protocol.statematcher(connection)
            while True:
                ev, m = await M_(conn_down, notification)
                if m is conn_down:
                    break
                else:
                    for buuid, v in ev.params[1]['Bridge'].items():
                        # If a bridge's name or datapath-id is changed, we remove this bridge and add it again
                        if 'old' in v:
                            # A bridge is deleted
                            bridges = self.managed_bridges[connection]
                            for i in range(0, len(bridges)):
                                if buuid == bridges[i][3]:
                                    self.scheduler.emergesend(OVSDBBridgeSetup(OVSDBBridgeSetup.DOWN,
                                                                               bridges[i][1],
                                                                               system_id,
                                                                               bridges[i][2],
                                                                               connection,
                                                                               connection.connmark,
                                                                               vhost,
                                                                               bridges[i][3],
                                                                               new_datapath_id =
                                                                                int(v['new']['datapath_id'], 16) if 'new' in v and 'datapath_id' in v['new']
                                                                                else None))
                                    del self.managed_conns[(vhost, bridges[i][1])]
                                    del bridges[i]
                                    break
                        if 'new' in v:
                            # A bridge is added
                            self.apiroutine.subroutine(self._update_bridge(connection, protocol, buuid, vhost))
        except JsonRPCProtocolException:
            pass
        finally:
            del connection._ovsdb_manager_get_bridges

    async def _manage_existing(self):
        conns = await call_api(self.apiroutine, "jsonrpcserver", "getconnections", {})
        vb = self.vhostbind
        for c in conns:
            if vb is None or c.protocol.vhost in vb:
                if not hasattr(c, '_ovsdb_manager_get_bridges'):
                    c._ovsdb_manager_get_bridges = self.apiroutine.subroutine(self._get_bridges(c, c.protocol))
        matchers = [OVSDBConnectionSetup.createMatcher(None, c, c.connmark) for c in conns
                    if vb is None or c.protocol.vhost in vb]
        await self.apiroutine.wait_for_all(*matchers)
        self._synchronized = True
        await self.apiroutine.wait_for_send(ModuleNotification(self.getServiceName(), 'synchronized'))

    async def _wait_for_sync(self):
        if not self._synchronized:
            await ModuleNotification.createMatcher(self.getServiceName(), 'synchronized')
    
    async def _manage_conns(self):
        try:
            self.apiroutine.subroutine(self._manage_existing())
            vb = self.vhostbind
            if vb is not None:
                conn_up = JsonRPCConnectionStateEvent.createMatcher(state = JsonRPCConnectionStateEvent.CONNECTION_UP,
                                                                     _ismatch = lambda x: x.createby.vhost in vb)
                conn_down = JsonRPCConnectionStateEvent.createMatcher(state = JsonRPCConnectionStateEvent.CONNECTION_DOWN,
                                                                     _ismatch = lambda x: x.createby.vhost in vb)
            else:
                conn_up = JsonRPCConnectionStateEvent.createMatcher(state = JsonRPCConnectionStateEvent.CONNECTION_UP)
                conn_down = JsonRPCConnectionStateEvent.createMatcher(state = JsonRPCConnectionStateEvent.CONNECTION_DOWN)
            while True:
                ev, m = await M_(conn_up, conn_down)
                if m is conn_up:
                    if not hasattr(ev.connection, '_ovsdb_manager_get_bridges'):
                        ev.connection._ovsdb_manager_get_bridges = self.apiroutine.subroutine(self._get_bridges(ev.connection, ev.createby))
                else:
                    conn = ev.connection
                    bridges = self.managed_bridges.get(conn)
                    if bridges is not None:
                        del self.managed_systemids[(ev.createby.vhost, conn.ovsdb_systemid)]
                        del self.managed_bridges[conn]
                        for vhost, dpid, name, buuid in bridges:
                            del self.managed_conns[(vhost, dpid)]
                            self.scheduler.emergesend(OVSDBBridgeSetup(OVSDBBridgeSetup.DOWN,
                                                                       dpid,
                                                                       conn.ovsdb_systemid,
                                                                       name,
                                                                       conn,
                                                                       conn.connmark,
                                                                       ev.createby.vhost,
                                                                       buuid))
                        econns = self.endpoint_conns.get(_get_endpoint(conn))
                        if econns is not None:
                            try:
                                econns.remove(conn)
                            except ValueError:
                                pass
        finally:
            for c in self.managed_bridges.keys():
                if hasattr(c, '_ovsdb_manager_get_bridges'):
                    c._ovsdb_manager_get_bridges.close()
                bridges = self.managed_bridges.get(c)
                if bridges is not None:
                    for vhost, dpid, name, buuid in bridges:
                        del self.managed_conns[(vhost, dpid)]
                        self.scheduler.emergesend(OVSDBBridgeSetup(OVSDBBridgeSetup.DOWN,
                                                                   dpid, 
                                                                   c.ovsdb_systemid, 
                                                                   name, 
                                                                   c, 
                                                                   c.connmark, 
                                                                   c.protocol.vhost,
                                                                   buuid))
    async def getconnection(self, datapathid, vhost = ''):
        "Get current connection of datapath"
        await self._wait_for_sync()
        return self.managed_conns.get((vhost, datapathid))

    async def waitconnection(self, datapathid, timeout = 30, vhost = ''):
        "Wait for a datapath connection"
        c = await self.getconnection(datapathid, vhost)
        if c is None:
            timeout_, ev, m = await self.apiroutine.wait_with_timeout(timeout,
                                            OVSDBBridgeSetup.createMatcher(
                                                    state = OVSDBBridgeSetup.UP,
                                                    datapathid = datapathid, vhost = vhost))
            if timeout_:
                raise ConnectionResetException('Datapath is not connected')
            return ev.connection
        else:
            return c

    async def getdatapathids(self, vhost = ''):
        "Get All datapath IDs"
        await self._wait_for_sync()
        return [k[1] for k in self.managed_conns.keys() if k[0] == vhost]

    async def getalldatapathids(self):
        "Get all datapath IDs from any vhost. Return ``(vhost, datapathid)`` pair."
        await self._wait_for_sync()
        return list(self.managed_conns.keys())

    async def getallconnections(self, vhost = ''):
        "Get all connections from vhost. If vhost is None, return all connections from any host"
        await self._wait_for_sync()
        if vhost is None:
            return list(self.managed_bridges.keys())
        else:
            return list(k for k in self.managed_bridges.keys() if k.protocol.vhost == vhost)

    async def getbridges(self, connection):
        "Get all ``(dpid, name, _uuid)`` tuple on this connection"
        await self._wait_for_sync()
        bridges = self.managed_bridges.get(connection)
        if bridges is not None:
            return [(dpid, name, buuid) for _, dpid, name, buuid in bridges]
        else:
            return None

    async def getallbridges(self, vhost = None):
        "Get all ``(dpid, name, _uuid)`` tuple for all connections, optionally filtered by vhost"
        await self._wait_for_sync()
        if vhost is not None:
            return [(dpid, name, buuid)
                    for c, bridges in self.managed_bridges.items()
                    if c.protocol.vhost == vhost
                    for _, dpid, name, buuid in bridges]
        else:
            return [(dpid, name, buuid)
                    for c, bridges in self.managed_bridges.items()
                    for _, dpid, name, buuid in bridges]

    async def getbridge(self, connection, name):
        "Get datapath ID on this connection with specified name"
        await self._wait_for_sync()
        bridges = self.managed_bridges.get(connection)
        if bridges is not None:
            for _, dpid, n, _ in bridges:
                if n == name:
                    return dpid
            return None
        else:
            return None

    async def waitbridge(self, connection, name, timeout = 30):
        "Wait for bridge with specified name appears and return the datapath-id"
        bnames = self.bridgenames
        if bnames is not None and name not in bnames:
            raise OVSDBBridgeNotAppearException('Bridge ' + repr(name) + ' does not appear: it is not in the selected bridge names')
        dpid = await self.getbridge(connection, name)
        if dpid is None:
            bridge_setup = OVSDBBridgeSetup.createMatcher(OVSDBBridgeSetup.UP,
                                                         None,
                                                         None,
                                                         name,
                                                         connection
                                                         )
            conn_down = JsonRPCConnectionStateEvent.createMatcher(JsonRPCConnectionStateEvent.CONNECTION_DOWN,
                                                                  connection,
                                                                  connection.connmark)
            timeout_, ev, m = await self.apiroutine.wait_with_timeout(timeout, bridge_setup, conn_down)
            if timeout_:
                raise OVSDBBridgeNotAppearException('Bridge ' + repr(name) + ' does not appear')
            elif m is conn_down:
                raise ConnectionResetException('Connection is down before bridge ' + repr(name) + ' appears')
            else:
                return ev.datapathid
        else:
            return dpid

    async def getbridgebyuuid(self, connection, uuid):
        "Get datapath ID of bridge on this connection with specified _uuid"
        await self._wait_for_sync()
        bridges = self.managed_bridges.get(connection)
        if bridges is not None:
            for _, dpid, _, buuid in bridges:
                if buuid == uuid:
                    return dpid
            return None
        else:
            return None

    async def waitbridgebyuuid(self, connection, uuid, timeout = 30):
        "Wait for bridge with specified _uuid appears and return the datapath-id"
        dpid = await self.getbridgebyuuid(connection, uuid)
        if dpid is None:
            bridge_setup = OVSDBBridgeSetup.createMatcher(state = OVSDBBridgeSetup.UP,
                                                         connection = connection,
                                                         bridgeuuid = uuid
                                                         )
            conn_down = JsonRPCConnectionStateEvent.createMatcher(JsonRPCConnectionStateEvent.CONNECTION_DOWN,
                                                                  connection,
                                                                  connection.connmark)
            timeout_, ev, m = await self.apiroutine.wait_with_timeout(timeout, bridge_setup, conn_down)
            if timeout_:
                raise OVSDBBridgeNotAppearException('Bridge ' + repr(uuid) + ' does not appear')
            elif m is conn_down:
                raise ConnectionResetException('Connection is down before bridge ' + repr(uuid) + ' appears')
            else:
                return ev.datapathid
        else:
            return dpid

    async def getsystemids(self, vhost = ''):
        "Get All system-ids"
        await self._wait_for_sync()
        return [k[1] for k in self.managed_systemids.keys() if k[0] == vhost]

    async def getallsystemids(self):
        "Get all system-ids from any vhost. Return ``(vhost, system-id)`` pair."
        await self._wait_for_sync()
        return list(self.managed_systemids.keys())

    async def getconnectionbysystemid(self, systemid, vhost = ''):
        await self._wait_for_sync()
        return self.managed_systemids.get((vhost, systemid))

    async def waitconnectionbysystemid(self, systemid, timeout = 30, vhost = ''):
        "Wait for a connection with specified system-id"
        c = await self.getconnectionbysystemid(systemid, vhost)
        if c is None:
            timeout_, ev, _ = await self.apiroutine.wait_with_timeout(
                                            timeout, 
                                            OVSDBConnectionSetup.createMatcher(
                                                    systemid, None, None, vhost))
            if timeout_:
                raise ConnectionResetException('Datapath is not connected')
            return ev.connection
        else:
            return c

    async def getconnectionsbyendpoint(self, endpoint, vhost = ''):
        "Get connection by endpoint address (IP, IPv6 or UNIX socket address)"
        await self._wait_for_sync()
        return self.endpoint_conns.get((vhost, endpoint))

    async def getconnectionsbyendpointname(self, name, vhost = '', timeout = 30):
        "Get connection by endpoint name (Domain name, IP or IPv6 address)"
        # Resolve the name
        if not name:
            endpoint = ''
            return await self.getconnectionbyendpoint(endpoint, vhost)
        else:
            request = (name, 0, socket.AF_UNSPEC, socket.SOCK_STREAM, socket.IPPROTO_TCP, socket.AI_ADDRCONFIG | socket.AI_V4MAPPED)
            # Resolve hostname
            await self.apiroutine.wait_for_send(ResolveRequestEvent(request))
            timeout_, ev, _ = await self.apiroutine.wait_with_timeout(timeout, ResolveResponseEvent.createMatcher(request))
            if timeout_:
                # Resolve is only allowed through asynchronous resolver
                #try:
                #    self.addrinfo = socket.getaddrinfo(self.hostname, self.port, socket.AF_UNSPEC, socket.SOCK_DGRAM if self.udp else socket.SOCK_STREAM, socket.IPPROTO_UDP if self.udp else socket.IPPROTO_TCP, socket.AI_ADDRCONFIG|socket.AI_NUMERICHOST)
                #except:
                raise IOError('Resolve hostname timeout: ' + name)
            else:
                if hasattr(ev, 'error'):
                    raise IOError('Cannot resolve hostname: ' + name)
                resp = ev.response
                for r in resp:
                    raddr = r[4]
                    if isinstance(raddr, tuple):
                        # Ignore port
                        endpoint = raddr[0]
                    else:
                        # Unix socket? This should not happen, but in case...
                        endpoint = raddr
                    r = await self.getconnectionsbyendpoint(endpoint, vhost)
                    if r is not None:
                        return r

    async def getendpoints(self, vhost = ''):
        "Get all endpoints for vhost"
        await self._wait_for_sync()
        return [k[1] for k in self.endpoint_conns if k[0] == vhost]

    async def getallendpoints(self):
        "Get all endpoints from any vhost. Return ``(vhost, endpoint)`` pairs."
        await self._wait_for_sync()
        return list(self.endpoint_conns.keys())

    async def getbridgeinfo(self, datapathid, vhost = ''):
        "Get ``(bridgename, systemid, bridge_uuid)`` tuple from bridge datapathid"
        c = await self.getconnection(datapathid, vhost)
        if c is not None:
            bridges = self.managed_bridges.get(c)
            if bridges is not None:
                for _, dpid, n, buuid in bridges:
                    if dpid == datapathid:
                        return (n, c.ovsdb_systemid, buuid)
                return None
            else:
                return None
        else:
            return None

    async def waitbridgeinfo(self, datapathid, timeout = 30, vhost = ''):
        "Wait for bridge with datapathid, and return ``(bridgename, systemid, bridge_uuid)`` tuple"
        bridge = await self.getbridgeinfo(datapathid, vhost)
        if bridge is None:
            timeout_, ev, m = await self.apiroutine.wait_with_timeout(
                                            timeout,
                                            OVSDBBridgeSetup.createMatcher(
                                                        OVSDBBridgeSetup.UP, datapathid,
                                                        None, None, None, None,
                                                        vhost))
            if timeout_:
                raise OVSDBBridgeNotAppearException('Bridge 0x%016x does not appear before timeout' % (datapathid,))
            return (ev.name, ev.systemid, ev.bridgeuuid)
        else:
            return bridge
